/*
**  SSEmem benchmarks for RAMspeed (amd64, malloc()-friendly)
**
**  (UNIX release)
**
**
**  This software is distributed under the terms of The Alasir Licence (TAL).
**  You should have received a copy of the licence together with the software.
**  If not, you should download it from http://www.alasir.com/licence/TAL.txt
**
**
**  Copyright (c) 2005 Paul V. Bolotoff <walter@alasir.com>
**
**  All rights reserved.
**
*/

/* NOTE: this code has been derived from the original SSEmem code, but it
 * contains work-arounds against badly written malloc() functions, which
 * return addresses not aligned at 16-byte boundaries, thus causing the
 * original code to fail with a segmentation fault most likely; the GNU C
 * library of different versions shows such an ignorant behaviour - PVB */

.globl ssecp
ssecp:
/* set up the stack frame */
	pushq	%rbx
	pushq	%rbp
	pushq	%r12
	pushq	%r13
	subq	$32, %rsp
/* process the parameters */
	movq	%rdi, %rbx
	movq	%rsi, %rbp
	shlq	$10, %rbx
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r12
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r13
/* prefill */
	movq	%rbx, %rcx
	shrq	$4, %rcx
	movq	%r12, %rdi
	finit
	fldpi
.precp:
	fst	0(%rdi)
	fst	4(%rdi)
	fst	8(%rdi)
	fst	12(%rdi)
	addq	$16, %rdi
	loopq	.precp
	ffree	%st(0)
/* wall time (start) */
	leaq	0(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* execute */
	shrq	$9, %rbx
	movq	%rbx, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	$512, %rax
.maincp:
	movups	0(%rdi), %xmm0
	movups	16(%rdi), %xmm1
	movups	32(%rdi), %xmm2
	movups	48(%rdi), %xmm3
	movups	64(%rdi), %xmm4
	movups	80(%rdi), %xmm5
	movups	96(%rdi), %xmm6
	movups	112(%rdi), %xmm7
	movups	%xmm0, 0(%rsi)
	movups	%xmm1, 16(%rsi)
	movups	%xmm2, 32(%rsi)
	movups	%xmm3, 48(%rsi)
	movups	%xmm4, 64(%rsi)
	movups	%xmm5, 80(%rsi)
	movups	%xmm6, 96(%rsi)
	movups	%xmm7, 112(%rsi)
	movups	128(%rdi), %xmm0
	movups	144(%rdi), %xmm1
	movups	160(%rdi), %xmm2
	movups	176(%rdi), %xmm3
	movups	192(%rdi), %xmm4
	movups	208(%rdi), %xmm5
	movups	224(%rdi), %xmm6
	movups	240(%rdi), %xmm7
	movups	%xmm0, 128(%rsi)
	movups	%xmm1, 144(%rsi)
	movups	%xmm2, 160(%rsi)
	movups	%xmm3, 176(%rsi)
	movups	%xmm4, 192(%rsi)
	movups	%xmm5, 208(%rsi)
	movups	%xmm6, 224(%rsi)
	movups	%xmm7, 240(%rsi)
	movups	256(%rdi), %xmm0
	movups	272(%rdi), %xmm1
	movups	288(%rdi), %xmm2
	movups	304(%rdi), %xmm3
	movups	320(%rdi), %xmm4
	movups	336(%rdi), %xmm5
	movups	352(%rdi), %xmm6
	movups	368(%rdi), %xmm7
	movups	%xmm0, 256(%rsi)
	movups	%xmm1, 272(%rsi)
	movups	%xmm2, 288(%rsi)
	movups	%xmm3, 304(%rsi)
	movups	%xmm4, 320(%rsi)
	movups	%xmm5, 336(%rsi)
	movups	%xmm6, 352(%rsi)
	movups	%xmm7, 368(%rsi)
	movups	384(%rdi), %xmm0
	movups	400(%rdi), %xmm1
	movups	416(%rdi), %xmm2
	movups	432(%rdi), %xmm3
	movups	448(%rdi), %xmm4
	movups	464(%rdi), %xmm5
	movups	480(%rdi), %xmm6
	movups	496(%rdi), %xmm7
	movups	%xmm0, 384(%rsi)
	movups	%xmm1, 400(%rsi)
	movups	%xmm2, 416(%rsi)
	movups	%xmm3, 432(%rsi)
	movups	%xmm4, 448(%rsi)
	movups	%xmm5, 464(%rsi)
	movups	%xmm6, 480(%rsi)
	movups	%xmm7, 496(%rsi)
	addq	%rax, %rdi
	addq	%rax, %rsi
	decq	%rcx
	jnz	.maincp

	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%rbx, %rcx
	decq	%rbp
	jnz	.maincp

/* wall time (finish) */
	leaq	16(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* free */
	movq	%r13, %rdi
	call	free
/* free */
	movq	%r12, %rdi
	call	free
/* calculate */
	movq	16(%rsp), %rax
	subq	0(%rsp), %rax
	movq	$1000000, %rdx
	mulq	%rdx
	addq	24(%rsp), %rax
	subq	8(%rsp), %rax
/* restore and return */
	addq	$32, %rsp
	popq	%r13
	popq	%r12
	popq	%rbp
	popq	%rbx
	ret

.globl ssesc
ssesc:
/* set up the stack frame */
	pushq	%rbx
	pushq	%rbp
	pushq	%r12
	pushq	%r13
	subq	$32, %rsp
/* process the parameters */
	movq	%rdi, %rbx
	movq	%rsi, %rbp
	shlq	$10, %rbx
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r12
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r13
/* prefill */
	movq	%rbx, %rcx
	shrq	$4, %rcx
	movq	%r12, %rdi
	finit
	fldpi
.presc:
	fst	0(%rdi)
	fst	4(%rdi)
	fst	8(%rdi)
	fst	12(%rdi)
	addq	$16, %rdi
	loopq	.presc
	ffree	%st(0)
/* wall time (start) */
	leaq	0(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* create the multiplier */
	fldln2
	fst	16(%rsp)
	fst	20(%rsp)
	movlps	16(%rsp), %xmm8
	movlhps	%xmm8, %xmm8
	ffree	%st(0)
/* execute */
	shrq	$9, %rbx
	movq	%rbx, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	$512, %rax
.mainsc:
	movups	0(%rdi), %xmm0
	movups	16(%rdi), %xmm1
	movups	32(%rdi), %xmm2
	movups	48(%rdi), %xmm3
	movups	64(%rdi), %xmm4
	movups	80(%rdi), %xmm5
	movups	96(%rdi), %xmm6
	movups	112(%rdi), %xmm7
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	mulps	%xmm8, %xmm4
	mulps	%xmm8, %xmm5
	mulps	%xmm8, %xmm6
	mulps	%xmm8, %xmm7
	movups	%xmm0, 0(%rsi)
	movups	%xmm1, 16(%rsi)
	movups	%xmm2, 32(%rsi)
	movups	%xmm3, 48(%rsi)
	movups	%xmm4, 64(%rsi)
	movups	%xmm5, 80(%rsi)
	movups	%xmm6, 96(%rsi)
	movups	%xmm7, 112(%rsi)
	movups	128(%rdi), %xmm0
	movups	144(%rdi), %xmm1
	movups	160(%rdi), %xmm2
	movups	176(%rdi), %xmm3
	movups	192(%rdi), %xmm4
	movups	208(%rdi), %xmm5
	movups	224(%rdi), %xmm6
	movups	240(%rdi), %xmm7
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	mulps	%xmm8, %xmm4
	mulps	%xmm8, %xmm5
	mulps	%xmm8, %xmm6
	mulps	%xmm8, %xmm7
	movups	%xmm0, 128(%rsi)
	movups	%xmm1, 144(%rsi)
	movups	%xmm2, 160(%rsi)
	movups	%xmm3, 176(%rsi)
	movups	%xmm4, 192(%rsi)
	movups	%xmm5, 208(%rsi)
	movups	%xmm6, 224(%rsi)
	movups	%xmm7, 240(%rsi)
	movups	256(%rdi), %xmm0
	movups	272(%rdi), %xmm1
	movups	288(%rdi), %xmm2
	movups	304(%rdi), %xmm3
	movups	320(%rdi), %xmm4
	movups	336(%rdi), %xmm5
	movups	352(%rdi), %xmm6
	movups	368(%rdi), %xmm7
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	mulps	%xmm8, %xmm4
	mulps	%xmm8, %xmm5
	mulps	%xmm8, %xmm6
	mulps	%xmm8, %xmm7
	movups	%xmm0, 256(%rsi)
	movups	%xmm1, 272(%rsi)
	movups	%xmm2, 288(%rsi)
	movups	%xmm3, 304(%rsi)
	movups	%xmm4, 320(%rsi)
	movups	%xmm5, 336(%rsi)
	movups	%xmm6, 352(%rsi)
	movups	%xmm7, 368(%rsi)
	movups	384(%rdi), %xmm0
	movups	400(%rdi), %xmm1
	movups	416(%rdi), %xmm2
	movups	432(%rdi), %xmm3
	movups	448(%rdi), %xmm4
	movups	464(%rdi), %xmm5
	movups	480(%rdi), %xmm6
	movups	496(%rdi), %xmm7
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	mulps	%xmm8, %xmm4
	mulps	%xmm8, %xmm5
	mulps	%xmm8, %xmm6
	mulps	%xmm8, %xmm7
	movups	%xmm0, 384(%rsi)
	movups	%xmm1, 400(%rsi)
	movups	%xmm2, 416(%rsi)
	movups	%xmm3, 432(%rsi)
	movups	%xmm4, 448(%rsi)
	movups	%xmm5, 464(%rsi)
	movups	%xmm6, 480(%rsi)
	movups	%xmm7, 496(%rsi)
	addq	%rax, %rdi
	addq	%rax, %rsi
	decq	%rcx
	jnz	.mainsc

	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%rbx, %rcx
	decq	%rbp
	jnz	.mainsc

/* wall time (finish) */
	leaq	16(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* free */
	movq	%r13, %rdi
	call	free
/* free */
	movq	%r12, %rdi
	call	free
/* calculate */
	movq	16(%rsp), %rax
	subq	0(%rsp), %rax
	movq	$1000000, %rdx
	mulq	%rdx
	addq	24(%rsp), %rax
	subq	8(%rsp), %rax
/* restore and return */
	addq	$32, %rsp
	popq	%r13
	popq	%r12
	popq	%rbp
	popq	%rbx
	ret

.globl ssead
ssead:
/* set up the stack frame */
	pushq	%rbx
	pushq	%rbp
	pushq	%r12
	pushq	%r13
	pushq	%r14
	pushq	%r15
	subq	$32, %rsp
/* process the parameters */
	movq	%rdi, %rbx
	movq	%rsi, %r15
	shlq	$10, %rbx
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r12
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r13
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r14
/* prefill */
	movq	%rbx, %rcx
	shrq	$4, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	finit
	fldpi
.pread:
	fst	0(%rdi)
	fst	4(%rdi)
	fst	8(%rdi)
	fst	12(%rdi)
	fst	0(%rsi)
	fst	4(%rsi)
	fst	8(%rsi)
	fst	12(%rsi)
	addq	$16, %rdi
	addq	$16, %rsi
	loopq	.pread
	ffree	%st(0)
/* wall time (start) */
	leaq	0(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* execute */
	shrq	$9, %rbx
	movq	%rbx, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%r14, %rbp
	movq	$512, %rax
.mainad:
	movups	0(%rdi), %xmm0
	movups	16(%rdi), %xmm1
	movups	32(%rdi), %xmm2
	movups	48(%rdi), %xmm3
	movups	64(%rdi), %xmm4
	movups	80(%rdi), %xmm5
	movups	96(%rdi), %xmm6
	movups	112(%rdi), %xmm7
	movups	0(%rsi), %xmm8
	movups	16(%rsi), %xmm9
	movups	32(%rsi), %xmm10
	movups	48(%rsi), %xmm11
	movups	64(%rsi), %xmm12
	movups	80(%rsi), %xmm13
	movups	96(%rsi), %xmm14
	movups	112(%rsi), %xmm15
	addps	%xmm8, %xmm0
	addps	%xmm9, %xmm1
	addps	%xmm10, %xmm2
	addps	%xmm11, %xmm3
	addps	%xmm12, %xmm4
	addps	%xmm13, %xmm5
	addps	%xmm14, %xmm6
	addps	%xmm15, %xmm7
	movups	%xmm0, 0(%rbp)
	movups	%xmm1, 16(%rbp)
	movups	%xmm2, 32(%rbp)
	movups	%xmm3, 48(%rbp)
	movups	%xmm4, 64(%rbp)
	movups	%xmm5, 80(%rbp)
	movups	%xmm6, 96(%rbp)
	movups	%xmm7, 112(%rbp)
	movups	128(%rdi), %xmm0
	movups	144(%rdi), %xmm1
	movups	160(%rdi), %xmm2
	movups	176(%rdi), %xmm3
	movups	192(%rdi), %xmm4
	movups	208(%rdi), %xmm5
	movups	224(%rdi), %xmm6
	movups	240(%rdi), %xmm7
	movups	128(%rsi), %xmm8
	movups	144(%rsi), %xmm9
	movups	160(%rsi), %xmm10
	movups	176(%rsi), %xmm11
	movups	192(%rsi), %xmm12
	movups	208(%rsi), %xmm13
	movups	224(%rsi), %xmm14
	movups	240(%rsi), %xmm15
	addps	%xmm8, %xmm0
	addps	%xmm9, %xmm1
	addps	%xmm10, %xmm2
	addps	%xmm11, %xmm3
	addps	%xmm12, %xmm4
	addps	%xmm13, %xmm5
	addps	%xmm14, %xmm6
	addps	%xmm15, %xmm7
	movups	%xmm0, 128(%rbp)
	movups	%xmm1, 144(%rbp)
	movups	%xmm2, 160(%rbp)
	movups	%xmm3, 176(%rbp)
	movups	%xmm4, 192(%rbp)
	movups	%xmm5, 208(%rbp)
	movups	%xmm6, 224(%rbp)
	movups	%xmm7, 240(%rbp)
	movups	256(%rdi), %xmm0
	movups	272(%rdi), %xmm1
	movups	288(%rdi), %xmm2
	movups	304(%rdi), %xmm3
	movups	320(%rdi), %xmm4
	movups	336(%rdi), %xmm5
	movups	352(%rdi), %xmm6
	movups	368(%rdi), %xmm7
	movups	256(%rsi), %xmm8
	movups	272(%rsi), %xmm9
	movups	288(%rsi), %xmm10
	movups	304(%rsi), %xmm11
	movups	320(%rsi), %xmm12
	movups	336(%rsi), %xmm13
	movups	352(%rsi), %xmm14
	movups	368(%rsi), %xmm15
	addps	%xmm8, %xmm0
	addps	%xmm9, %xmm1
	addps	%xmm10, %xmm2
	addps	%xmm11, %xmm3
	addps	%xmm12, %xmm4
	addps	%xmm13, %xmm5
	addps	%xmm14, %xmm6
	addps	%xmm15, %xmm7
	movups	%xmm0, 256(%rbp)
	movups	%xmm1, 272(%rbp)
	movups	%xmm2, 288(%rbp)
	movups	%xmm3, 304(%rbp)
	movups	%xmm4, 320(%rbp)
	movups	%xmm5, 336(%rbp)
	movups	%xmm6, 352(%rbp)
	movups	%xmm7, 368(%rbp)
	movups	384(%rdi), %xmm0
	movups	400(%rdi), %xmm1
	movups	416(%rdi), %xmm2
	movups	432(%rdi), %xmm3
	movups	448(%rdi), %xmm4
	movups	464(%rdi), %xmm5
	movups	480(%rdi), %xmm6
	movups	496(%rdi), %xmm7
	movups	384(%rsi), %xmm8
	movups	400(%rsi), %xmm9
	movups	416(%rsi), %xmm10
	movups	432(%rsi), %xmm11
	movups	448(%rsi), %xmm12
	movups	464(%rsi), %xmm13
	movups	480(%rsi), %xmm14
	movups	496(%rsi), %xmm15
	addps	%xmm8, %xmm0
	addps	%xmm9, %xmm1
	addps	%xmm10, %xmm2
	addps	%xmm11, %xmm3
	addps	%xmm12, %xmm4
	addps	%xmm13, %xmm5
	addps	%xmm14, %xmm6
	addps	%xmm15, %xmm7
	movups	%xmm0, 384(%rbp)
	movups	%xmm1, 400(%rbp)
	movups	%xmm2, 416(%rbp)
	movups	%xmm3, 432(%rbp)
	movups	%xmm4, 448(%rbp)
	movups	%xmm5, 464(%rbp)
	movups	%xmm6, 480(%rbp)
	movups	%xmm7, 496(%rbp)
	addq	%rax, %rdi
	addq	%rax, %rsi
	addq	%rax, %rbp
	decq	%rcx
	jnz	.mainad

	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%r14, %rbp
	movq	%rbx, %rcx
	decq	%r15
	jnz	.mainad

/* wall time (finish) */
	leaq	16(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* free */
	movq	%r14, %rdi
	call	free
/* free */
	movq	%r13, %rdi
	call	free
/* free */
	movq	%r12, %rdi
	call	free
/* calculate */
	movq	16(%rsp), %rax
	subq	0(%rsp), %rax
	movq	$1000000, %rdx
	mulq	%rdx
	addq	24(%rsp), %rax
	subq	8(%rsp), %rax
/* restore and return */
	addq	$32, %rsp
	popq	%r15
	popq	%r14
	popq	%r13
	popq	%r12
	popq	%rbp
	popq	%rbx
	ret

.globl ssetr
ssetr:
/* set up the stack frame */
	pushq	%rbx
	pushq	%rbp
	pushq	%r12
	pushq	%r13
	pushq	%r14
	pushq	%r15
	subq	$32, %rsp
/* process the parameters */
	movq	%rdi, %rbx
	movq	%rsi, %r15
	shlq	$10, %rbx
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r12
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r13
/* allocate */
	movq	%rbx, %rdi
	call	malloc
	movq	%rax, %r14
/* prefill */
	movq	%rbx, %rcx
	shrq	$4, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	finit
	fldpi
.pretr:
	fst	0(%rdi)
	fst	4(%rdi)
	fst	8(%rdi)
	fst	12(%rdi)
	fst	0(%rsi)
	fst	4(%rsi)
	fst	8(%rsi)
	fst	12(%rsi)
	addq	$16, %rdi
	addq	$16, %rsi
	loopq	.pretr
	ffree	%st(0)
/* wall time (start) */
	leaq	0(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* create the multiplier */
	fldln2
	fst	16(%rsp)
	fst	20(%rsp)
	movlps	16(%rsp), %xmm8
	movlhps	%xmm8, %xmm8
	ffree	%st(0)
/* execute */
	shrq	$9, %rbx
	movq	%rbx, %rcx
	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%r14, %rbp
	movq	$512, %rax
.maintr:
	movups	0(%rsi), %xmm0
	movups	16(%rsi), %xmm1
	movups	32(%rsi), %xmm2
	movups	48(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	0(%rdi), %xmm4
	movups	16(%rdi), %xmm5
	movups	32(%rdi), %xmm6
	movups	48(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 0(%rbp)
	movups	%xmm1, 16(%rbp)
	movups	%xmm2, 32(%rbp)
	movups	%xmm3, 48(%rbp)
	movups	64(%rsi), %xmm0
	movups	80(%rsi), %xmm1
	movups	96(%rsi), %xmm2
	movups	112(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	64(%rdi), %xmm4
	movups	80(%rdi), %xmm5
	movups	96(%rdi), %xmm6
	movups	112(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 64(%rbp)
	movups	%xmm1, 80(%rbp)
	movups	%xmm2, 96(%rbp)
	movups	%xmm3, 112(%rbp)
	movups	128(%rsi), %xmm0
	movups	144(%rsi), %xmm1
	movups	160(%rsi), %xmm2
	movups	176(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	128(%rdi), %xmm4
	movups	144(%rdi), %xmm5
	movups	160(%rdi), %xmm6
	movups	176(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 128(%rbp)
	movups	%xmm1, 144(%rbp)
	movups	%xmm2, 160(%rbp)
	movups	%xmm3, 176(%rbp)
	movups	192(%rsi), %xmm0
	movups	208(%rsi), %xmm1
	movups	224(%rsi), %xmm2
	movups	240(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	192(%rdi), %xmm4
	movups	208(%rdi), %xmm5
	movups	224(%rdi), %xmm6
	movups	240(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 192(%rbp)
	movups	%xmm1, 208(%rbp)
	movups	%xmm2, 224(%rbp)
	movups	%xmm3, 240(%rbp)
	movups	256(%rsi), %xmm0
	movups	272(%rsi), %xmm1
	movups	288(%rsi), %xmm2
	movups	304(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	256(%rdi), %xmm4
	movups	272(%rdi), %xmm5
	movups	288(%rdi), %xmm6
	movups	304(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 256(%rbp)
	movups	%xmm1, 272(%rbp)
	movups	%xmm2, 288(%rbp)
	movups	%xmm3, 304(%rbp)
	movups	320(%rsi), %xmm0
	movups	336(%rsi), %xmm1
	movups	352(%rsi), %xmm2
	movups	368(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	320(%rdi), %xmm4
	movups	336(%rdi), %xmm5
	movups	352(%rdi), %xmm6
	movups	368(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 320(%rbp)
	movups	%xmm1, 336(%rbp)
	movups	%xmm2, 352(%rbp)
	movups	%xmm3, 368(%rbp)
	movups	384(%rsi), %xmm0
	movups	400(%rsi), %xmm1
	movups	416(%rsi), %xmm2
	movups	432(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	384(%rdi), %xmm4
	movups	400(%rdi), %xmm5
	movups	416(%rdi), %xmm6
	movups	432(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 384(%rbp)
	movups	%xmm1, 400(%rbp)
	movups	%xmm2, 416(%rbp)
	movups	%xmm3, 432(%rbp)
	movups	448(%rsi), %xmm0
	movups	464(%rsi), %xmm1
	movups	480(%rsi), %xmm2
	movups	496(%rsi), %xmm3
	mulps	%xmm8, %xmm0
	mulps	%xmm8, %xmm1
	mulps	%xmm8, %xmm2
	mulps	%xmm8, %xmm3
	movups	448(%rdi), %xmm4
	movups	464(%rdi), %xmm5
	movups	480(%rdi), %xmm6
	movups	496(%rdi), %xmm7
	addps	%xmm4, %xmm0
	addps	%xmm5, %xmm1
	addps	%xmm6, %xmm2
	addps	%xmm7, %xmm3
	movups	%xmm0, 448(%rbp)
	movups	%xmm1, 464(%rbp)
	movups	%xmm2, 480(%rbp)
	movups	%xmm3, 496(%rbp)
	addq	%rax, %rdi
	addq	%rax, %rsi
	addq	%rax, %rbp
	decq	%rcx
	jnz	.maintr

	movq	%r12, %rdi
	movq	%r13, %rsi
	movq	%r14, %rbp
	movq	%rbx, %rcx
	decq	%r15
	jnz	.maintr

/* wall time (finish) */
	leaq	16(%rsp), %rdi
	xorq	%rsi, %rsi
	call	gettimeofday
/* free */
	movq	%r14, %rdi
	call	free
/* free */
	movq	%r13, %rdi
	call	free
/* free */
	movq	%r12, %rdi
	call	free
/* calculate */
	movq	16(%rsp), %rax
	subq	0(%rsp), %rax
	movq	$1000000, %rdx
	mulq	%rdx
	addq	24(%rsp), %rax
	subq	8(%rsp), %rax
/* restore and return */
	addq	$32, %rsp
	popq	%r15
	popq	%r14
	popq	%r13
	popq	%r12
	popq	%rbp
	popq	%rbx
	ret
