/*
**  CPUinfo -- a CPU information retrieving tool (i386)
**
**
**  This software is distributed under the terms of The Alasir Licence (TAL).
**  You should have received a copy of the licence together with the software.
**  If not, you should download it from http://www.alasir.com/licence/TAL.txt
**
**
**  Copyright (c) 2004-05 Rhett M. Hollander <rhett@alasir.com>
**  Copyright (c) 2005 Paul V. Bolotoff <walter@alasir.com>
**
**  All rights reserved.
**
*/


/* KNOWN ISSUES:
 * 1) IBM BlueLightning CPUs are not recognised and misdetected as Cyrix CPUs;
 * 2) both Pentium II Xeon and Pentium III Xeon CPUs with 512Kb of S-cache are
 *    identified as Pentium II and Pentium III CPUs respectively, thanks to
 *    Intel;
 * 3) both AMD Athlon XP and MP are identified as just Athlon;
 * 4) some Intel Xeon and Xeon MP CPUs of the Prescott family are detected as
 *    Pentium 4 CPUs due to Intel's hardware mess;
 * 5) AMD Sempron (Socket A) CPUs are detected as Athlon (Thoroughbred, Barton,
 *    Thorton cores), because in fact they are. */


/* input: a bitmask of operating mode to switch to:
 * bit  0 -- verbose (1) or silent (0) mode
 * bit  1 -- privileged (1) or regular (0) mode
 * output: a bitmask of processor's features:
 * bit  0 -- FPU
 * bit  1 -- CMOV
 * bit  2 -- MMX
 * bit  3 -- Extended MMX (MMX+)
 * bit  4 -- 3DNow!
 * bit  5 -- Extended 3DNow! (3DNow!+)
 * bit  6 -- SSE
 * bit  7 -- SSE2
 * bit  8 -- SSE3
 * bits 9 to 13 are reserved
 * bits 14 to 17 -- vendor's code:
 *   0x0 -- unknown
 *   0x1 -- Intel
 *   0x2 -- AMD
 *   0x3 -- Cyrix
 *   0x4 -- Centaur
 *   0x5 -- Transmeta
 *   0x6 -- Rise
 *   0x7 -- NexGen
 *   0x8 -- UMC
 *   0x9 -- NSC
 *   0xA -- SiS
 * bits 18 to 21 -- CPU family
 * bits 22 to 25 -- CPU model
 * bits 26 to 29 -- CPU stepping
 * bit 30 -- verbose/silent mode sign (internal use only, 0 in output)
 * bit 31 -- privileged/regular mode sign (internal use only, 0 in output) */
.globl cpuinfo
cpuinfo:
	pushl	%ebx
	pushl	%ebp
	pushl	%esi
	pushl	%edi
	movl	%esp, %ebp
	xorl	%edi, %edi

/* set up mode bits */
	movl	20(%esp), %eax
	shll	$30, %eax
	orl	%eax, %edi

/* push all printf() strings to the stack if in verbose mode */
	testl	$0x40000000, %edi
	jz	.nostrings
/* "%s family %X model %X stepping %X\n" is at -36(%ebp) */
	pushl	$0x00000A58
	pushl	$0x2520676E
	pushl	$0x69707065
	pushl	$0x74732058
	pushl	$0x25206C65
	pushl	$0x646F6D20
	pushl	$0x58252079
	pushl	$0x6C696D61
	pushl	$0x66207325
/* "model %X stepping %X revision %X\n" is at -70(%ebp) */
	pushw	$0x000A
	pushl	$0x5825206E
	pushl	$0x6F697369
	pushl	$0x76657220
	pushl	$0x58252067
	pushl	$0x6E697070
	pushl	$0x65747320
	pushl	$0x5825206C
	pushl	$0x65646F6D
/* "%s processor" is at -84(%ebp) */
	pushw	$0x0000
	pushl	$0x726F7373
	pushl	$0x65636F72
	pushl	$0x70207325
/* " %u.%uMHz\n" is at -96(%ebp) */
	pushl	$0x00000A7A
	pushl	$0x484D7525
	pushl	$0x2E752520
/* "\n" is at -98(%ebp) */
	pushw	$0x000A
/* "I-cache: %uKuops, %u-way\n" is at -124(%ebp) */
	pushw	$0x000A
	pushl	$0x7961772D
	pushl	$0x7525202C
	pushl	$0x73706F75
	pushl	$0x4B752520
	pushl	$0x3A656863
	pushl	$0x61632D49
/* "%c-cache: %uKb, %u-way, %u bytes per line\n" is at -168(%ebp) */
	pushl	$0x00000A65
	pushl	$0x6E696C20
	pushl	$0x72657020
	pushl	$0x73657479
	pushl	$0x62207525
	pushl	$0x202C7961
	pushl	$0x772D7525
	pushl	$0x202C624B
	pushl	$0x7525203A
	pushl	$0x65686361
	pushl	$0x632D6325
/* "%c-cache: [unknown]\n" is at -190(%ebp) */
	pushw	$0x0000
	pushl	$0x0A5D6E77
	pushl	$0x6F6E6B6E
	pushl	$0x755B203A
	pushl	$0x65686361
	pushl	$0x632D6325
/* "%c-cache: %uKb, %u-way, %u lines per tag, %u bytes per line\n" is at -252(%ebp) */
	pushw	$0x0000
	pushl	$0x0A656E69
	pushl	$0x6C207265
	pushl	$0x70207365
	pushl	$0x74796220
	pushl	$0x7525202C
	pushl	$0x67617420
	pushl	$0x72657020
	pushl	$0x73656E69
	pushl	$0x6C207525
	pushl	$0x202C7961
	pushl	$0x772D7525
	pushl	$0x202C624B
	pushl	$0x7525203A
	pushl	$0x65686361
	pushl	$0x632D6325
/* "%c-cache TLB: %u entries, %u-way, 4Kb page size\n" is at -302(%ebp) */
	pushw	$0x0000
	pushl	$0x0A657A69
	pushl	$0x73206567
	pushl	$0x61702062
	pushl	$0x4B34202C
	pushl	$0x7961772D
	pushl	$0x7525202C
	pushl	$0x73656972
	pushl	$0x746E6520
	pushl	$0x7525203A
	pushl	$0x424C5420
	pushl	$0x65686361
	pushl	$0x632D6325
/* "%c-cache TLB: [unknown]\n" is at -328(%ebp) */
	pushw	$0x0000
	pushl	$0x0A5D6E77
	pushl	$0x6F6E6B6E
	pushl	$0x755B203A
	pushl	$0x424C5420
	pushl	$0x65686361
	pushl	$0x632D6325
/* "Features:%s\n" is at -342(%ebp) */
	pushw	$0x0000
	pushl	$0x0A73253A
	pushl	$0x73657275
	pushl	$0x74616546

.nostrings:
/* standard function 0 (eflags, bit 21) */
	pushfl
	popl	%eax
	movl	%eax, %ebx
	xorl	$0x00200000, %eax
	pushl	%eax
	popfl
	pushfl
	popl	%eax
	cmpl	%eax, %ebx
	je	.nocpuid
	pushl	%ebx
	popfl

/* standard features;
 * NOTE: all CPUs supporting CPUID are supposed to have standard function 1
 *       implemented at least */
/* FPU (standard function 1, bit 0 of %edx) */
	movl	$1, %eax
	cpuid
	testl	$0x00000001, %edx
	jz	.stdinfo
	orw	$0x0001, %di
/* CMOV (standard function 1, bit 15 of %edx) */
	testl	$0x00008000, %edx
	jz	.mmcheck
	orw	$0x0002, %di
/* MMX (standard function 1, bit 23 of %edx) */
.mmcheck:
	testl	$0x00800000, %edx
	jz	.stdinfo
	orw 	$0x0004, %di
/* SSE (standard function 1, bit 25 of %edx) */
	testl	$0x02000000, %edx
	jz	.extcheck
	orw	$0x0040, %di
/* SSE implies MMX+ as well */
	orw	$0x0008, %di
/* SSE2 (standard function 1, bit 26 of %edx) */
	testl	$0x04000000, %edx
	jz	.extcheck
	orw	$0x0080, %di
/* SSE3 (standard function 1, bit 0 of %ecx) */
	testl	$0x00000001, %ecx
	jz	.extcheck
	orw	$0x0100, %di

/* extended features;
 * NOTE: extended function 0 returns the number of maximal supported extended
 *       function in %eax */
.extcheck:
	movl	$0x80000000, %eax
	cpuid
	cmpb	$0, %al
	je	.stdinfo
/* MMX+ (extended function 1, bit 22 of %edx) */
	movl	$0x80000001, %eax
	cpuid
	testl	$0x00400000, %edx
	jz	.tridcheck
	orw	$0x0008, %di
/* 3DNow! (extended function 1, bit 31 of %edx) */
.tridcheck:
	testl	$0x80000000, %edx
	jz	.stdinfo
	orw	$0x0010, %di
/* 3DNow!+ (extended function 1, bit 30 of %edx) */
	testl	$0x40000000, %edx
	jz	.stdinfo
	orw	$0x0020, %di

/* standard information */
/* recognise a vendor by its ID string (scan the last double-word only) */
.stdinfo:
	xorl	%eax, %eax
	cpuid
/* "GenuineIntel" */
	cmpl	$0x6C65746E, %ecx
	jne	.vend2
	orl	$0x00004000, %edi
/* "AuthenticAMD" */
.vend2:
	cmpl	$0x444D4163, %ecx
	jne	.vend3
	orl	$0x00008000, %edi
/* "CyrixInstead" */
.vend3:
	cmpl	$0x64616574, %ecx
	jne	.vend4
	orl	$0x0000C000, %edi
/* "CentaurHauls" */
.vend4:
	cmpl	$0x736C7561, %ecx
	jne	.vend5
	orl	$0x00010000, %edi
/* "GenuineTMx86" */
.vend5:
	cmpl	$0x3638784D, %ecx
	jne	.vend6
	orl	$0x00014000, %edi
/* "RiseRiseRise" */
.vend6:
	cmpl	$0x65736952, %ecx
	jne	.vend7
	orl	$0x00018000, %edi
/* "NexGenDriven" */
.vend7:
	cmpl	$0x6E657669, %ecx
	jne	.vend8
	orl	$0x0001C000, %edi
/* "UMC UMC UMC " */
.vend8:
	cmpl	$0x20434D55, %ecx
	jne	.vend9
	orl	$0x00020000, %edi
/* "Geode by NSC" */
.vend9:
	cmpl	$0x43534E20, %ecx
	jne	.vend10
	orl	$0x00024000, %edi
/* "SiS SiS SiS " */
.vend10:
	cmpl	$0x20536953, %ecx
	jne	.stdfms
	orl	$0x00028000, %edi

/* obtain family, model, stepping */
.stdfms:
	movl	$1, %eax
	cpuid
	movl	%eax, %ebx
	andl	$0x00000F00, %ebx
	shll	$10, %ebx
	orl	%ebx, %edi
	movl	%eax, %ecx
	andl	$0x000000F0, %ecx
	shll	$18, %ecx
	orl	%ecx, %edi
	movl	%eax, %edx
	andl	$0x0000000F, %edx
	shll	$26, %edx
	orl	%edx, %edi
/* print family, model, stepping, and vendor's ID */
	testl	$0x40000000, %edi
	jz	.capquit
	shrl	$18, %ebx
	shrl	$22, %ecx
	shrl	$26, %edx
	movl	%edx, -18(%esp)
	movl	%ecx, -22(%esp)
	movl	%ebx, -26(%esp)
	xorl	%eax, %eax
	cpuid
	pushw	$0x0000
	pushl	%ecx
	pushl	%edx
	pushl	%ebx
	movl	%esp, %eax
	subl	$12, %esp
	pushl	%eax
	leal	-36(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cputable

/* if CPUID is supported not */
.nocpuid:
/* detect either 386 or 486 (eflags, bit 18) */
	pushfl
	popl	%eax
	movl	%eax, %ebx
	xorl	$0x00040000, %eax
	pushl	%eax
	popfl
	pushfl
	popl	%eax
	cmpl	%eax, %ebx
	jne	.un486

/* a 386-class CPU has been detected */
	orl	$0x000C0000, %edi
/* try to detect a NexGen 586;
 * ZF shouldn't change after dividing 5 by 2
 * (from "Nx586 Processor Recognition Application Note") */
	xorl	%eax, %eax
	movb	$5, %al
	movb	$2, %bl
	divb	%bl
	jz	.nx586
/* generic 386 */
	testl	$0x40000000, %edi
	jz	.capquit
	pushw	$0x0000
	pushl	$0x36383320
	pushl	$0x63697265
	pushl	$0x6E65670A
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

.nx586:
	orl	$0x0001C000, %edi
	testl	$0x40000000, %edi
	jz	.capquit
.nx586_print:
	pushl	$0x00753434
	pushl	$0x2E302050
	pushl	$0x46363835
	pushl	$0x20726F20
	pushl	$0x7530352E
	pushl	$0x30203638
	pushl	$0x35206E65
	pushl	$0x4778654E
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* TSC is unsupported, don't know how to calculate */
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* 486-class CPU detected */
.un486:
	pushl	%ebx
	popfl
	orl	$0x00100000, %edi
/* try to detect a Cyrix CPU;
 * undefined flags must remain unchanged after dividing 5 by 2
 * (from "Cyrix CPU Detection Guide")
 * NOTE: this approach will also catch IBM Blue Lightning CPUs */
	xorl	%eax, %eax
	sahf
	movb	$5, %al
	movb	$2, %bl
	divb	%bl
	lahf
	cmpb	$2, %ah
	je	.cx486
/* generic 486 */
	testl	$0x40000000, %edi
	jz	.capquit
	pushw	$0x0000
	pushl	$0x36383420
	pushl	$0x63697265
	pushl	$0x6E65670A
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

.cx486:
	orl	$0x0000C000, %edi
	testl	$0x40000000, %edi
	jz	.capquit
	testl	$0x80000000, %edi
	jnz	.cxdirget
	pushl	$0x00003638
	pushl	$0x34207869
	pushl	$0x7279430A
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

.cxdirget:
/* read the DIR0 and DIR1 */
	xorl	%eax, %eax
	movb	$0xFE, %al
	outb	%al, $0x22
	inb	$0x23, %al
	movb	%al, %bl
	movb	$0xFF, %al
	outb	%al, $0x22
	inb	$0x23, %al
	movb	%al, %bh
/* print model, stepping, revision */
	andb	$0x0F, %al
	pushl	%eax
	movb	%bh, %al
	shrb	$4, %al
	pushl	%eax
	movb	%bl, %al
	pushl	%eax
	leal	-70(%ebp), %eax
	pushl	%eax
	call	printf
/* Cyrix look-up table */
	cmpb	$0x00, %bl
	je	.cx486slc
	cmpb	$0x01, %bl
	je	.cx486dlc
	cmpb	$0x02, %bl
	je	.cx486slc2
	cmpb	$0x03, %bl
	je	.cx486dlc2
	cmpb	$0x04, %bl
	je	.cx486srx
	cmpb	$0x05, %bl
	je	.cx486drx
	cmpb	$0x06, %bl
	je	.cx486srx2
	cmpb	$0x07, %bl
	je	.cx486drx2
	cmpb	$0x08, %bl
	je	.cx486sru
	cmpb	$0x09, %bl
	je	.cx486dru
	cmpb	$0x0A, %bl
	je	.cx486sru2
	cmpb	$0x0B, %bl
	je	.cx486dru2
	cmpb	$0x10, %bl
	je	.cx486s
	cmpb	$0x11, %bl
	je	.cx486s2
	cmpb	$0x12, %bl
	je	.cx486se
	cmpb	$0x13, %bl
	je	.cx486s2e
	cmpb	$0x1A, %bl
	je	.cx486dx
	cmpb	$0x1B, %bl
	je	.cx486dx2
	cmpb	$0x1F, %bl
	je	.cx486dx4
	movb	%bl, %al
	shrb	$4, %al
	cmpb	$0x02, %al
	je	.cxm1sc
	cmpb	$0x03, %al
	je	.cxm1_family
	cmpb	$0x04, %al
	je	.cxgx_family
	cmpb	$0x05, %al
	je	.cxm2
	xorw	%si, %si
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.cxquit

.cx486slc:
	movw	$1, %si
	pushl	$0x0000434C
	pushl	$0x53363834
	jmp	.cxquit

.cx486dlc:
	movw	$1, %si
	pushl	$0x0000434C
	pushl	$0x44363834
	jmp	.cxquit

.cx486slc2:
	movw	$1, %si
	pushl	$0x0032434C
	pushl	$0x53363834
	jmp	.cxquit

.cx486dlc2:
	movw	$1, %si
	pushl	$0x0032434C
	pushl	$0x44363834
	jmp	.cxquit

.cx486srx:
	movw	$1, %si
	pushl	$0x00007852
	pushl	$0x53363834
	jmp	.cxquit

.cx486drx:
	movw	$1, %si
	pushl	$0x00007852
	pushl	$0x44363834
	jmp	.cxquit
	
.cx486srx2:
	movw	$1, %si
	pushl	$0x00327852
	pushl	$0x53363834
	jmp	.cxquit

.cx486drx2:
	movw	$1, %si
	pushl	$0x00327852
	pushl	$0x44363834
	jmp	.cxquit

.cx486sru:
	movw	$1, %si
	pushl	$0x00007552
	pushl	$0x44363834
	jmp	.cxquit

.cx486dru:
	movw	$1, %si
	pushl	$0x00007552
	pushl	$0x44363834
	jmp	.cxquit

.cx486sru2:
	movw	$1, %si
	pushl	$0x00327552
	pushl	$0x44363834
	jmp	.cxquit

.cx486dru2:
	movw	$1, %si
	pushl	$0x00327552
	pushl	$0x44363834
	jmp	.cxquit

.cx486s:
	movw	$1, %si
	pushw	$0x0000
	pushl	$0x53363834
	jmp	.cxquit

.cx486s2:
	movw	$1, %si
	pushw 	$0x0032
	pushl	$0x53363834
	jmp	.cxquit

.cx486se:
	movw	$1, %si
	pushw	$0x0065
	pushl	$0x53363834
	jmp	.cxquit

.cx486s2e:
	movw	$1, %si
	pushl	$0x00006532
	pushl	$0x53363834
	jmp	.cxquit

.cx486dx:
	movw	$1, %si
	pushw	$0x0058
	pushl	$0x44363834
	orw	$0x0001, %di
	jmp	.cxquit

.cx486dx2:
	movw	$1, %si
	pushl	$0x00003258
	pushl	$0x44363834
	orw	$0x0001, %di
	jmp	.cxquit

.cx486dx4:
	movw	$1, %si
	pushl	$0x00003458
	pushl	$0x44363834
	orw	$0x0001, %di
	jmp	.cxquit

.cxm1sc:
	movw	$2, %si
	pushw	$0x0075
	pushl	$0x35362E30
	pushl	$0x20296373
	pushl	$0x314D2820
	pushl	$0x36387835
	orw	$0x0001, %di
	jmp	.cxquit

.cxm1_family:
	movw	$3, %si
	cmpb	$0x20, %bh
	jae	.cxm1l
.cxm1:
/* 150MHz version, at least, was 0.44u */
	pushw	$0x00075
	pushl	$0x34342E30
	pushl	$0x2D753536
	pushl	$0x2E302029
	pushl	$0x314D2820
	pushl	$0x36387836
	orw	$0x0001, %di
	jmp	.cxquit

.cxm1l:
	pushw	$0x0075
	pushl	$0x35332E30
	pushl	$0x20294C31
	pushl	$0x4D28204C
	pushl	$0x36387836
	orw	$0x0001, %di
	jmp	.cxquit

.cxgx_family:
	movw	$2, %si
	cmpb	$0x30, %bh
	jae	.cxgxm
.cxmgx:
	pushw	$0x0075
	pushl	$0x35332E30
	pushl	$0x20584761
	pushl	$0x6964654D
	orw	$0x0001, %di
	jmp	.cxquit
.cxgxm:
/* GXm was 0.35u, GXlv -- 0.25u */
	pushl	$0x00753532
	pushl	$0x2E302076
	pushl	$0x6C584720
	pushl	$0x726F2075
	pushl	$0x35332E30
	pushl	$0x206D5847
	orw	$0x0005, %di
	jmp	.cxquit

.cxm2:
	movw	$3, %si
/* was either 0.35u, or 0.30u, or 0.25u, or 0.18u */
	pushl	$0x00753831
	pushl	$0x2E302D75
	pushl	$0x35332E30
	pushl	$0x2029324D
	pushl	$0x2820584D
	pushl	$0x36387836
	orw	$0x0005, %di
	jmp	.cxquit

.cxquit:
	pushl	$0x20786972
	pushw	$0x7943
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
	testl	$0x80000000, %edi
	jnz	.cxfreq
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

.cxfreq:
/* obtain and print processor's frequency
 * (thanks to Cyrix for the information) */
	pushw	%si
	call	calcnt
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.nocxfreq
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.nocxfreq:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* CPU look-up table;
 * prints vendor's name, defines and prints CPU's trade-name and code-name
 * (if known), also technical process (if known for sure, in microns) */
.cputable:
	movl	%edi, %ebx
	movl	%edi, %eax
	shrl	$18, %ebx
	shrl	$14, %eax
	andw	$0x0FFF, %bx
	andb	$0x0F, %al
	cmpb	$0x01, %al
	je	.intel
	cmpb	$0x02, %al
	je	.amd
	cmpb	$0x03, %al
	je	.cyrix
	cmpb	$0x04, %al
	je	.centaur
	cmpb	$0x05, %al
	je	.transmeta
	cmpb	$0x06, %al
	je	.rise
	cmpb	$0x07, %al
	je	.nexgen
	cmpb	$0x08, %al
	je	.umc
	cmpb	$0x09, %al
	je	.nsc
	cmpb	$0x0A, %al
	je	.sis
	jmp	.cpufeat

/* Intel CPUs */
.intel:
	movl	%ebx, %eax
	andb	$0x0F, %al
	cmpb	$0x03, %al
	je	.i386
	cmpb	$0x04, %al
	je	.i486
	cmpb	$0x05, %al
	je	.i586
	cmpb	$0x06, %al
	je	.i686
	cmpb	$0x0F, %al
	je	.i786
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.iquit

.i386:
	pushl	$0x00363833
	jmp	.iquit

.i486:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.i486dx
	cmpb	$0x01, %al
	je	.i486dxa
	cmpb	$0x02, %al
	je	.i486sx
	cmpb	$0x03, %al
	je	.i486dx2wt
	cmpb	$0x04, %al
	je	.i486sl
	cmpb	$0x05, %al
	je	.i486sx2
	cmpb	$0x07, %al
	je	.i486dx2wb
	cmpb	$0x08, %al
	je	.i486dx4wt
	cmpb	$0x09, %al
	je	.i486dx4wb
	pushl	$0x005D7373
	pushl	$0x616C6320
	pushl	$0x3638345B
	jmp	.iquit

.i486dx:
	pushl	$0x00007530
	pushl	$0x2E312058
	pushl	$0x44363834
	jmp	.iquit

.i486dxa:
	pushl	$0x00007538
	pushl	$0x2E302058
	pushl	$0x44363834
	jmp	.iquit

.i486sx:
	pushl	$0x00007538
	pushl	$0x2E302058
	pushl	$0x53363834
	jmp	.iquit

.i486dx2wt:
	pushl	$0x0075382E
	pushl	$0x30202968
	pushl	$0x67756F72
	pushl	$0x68742D65
	pushl	$0x74697277
	pushl	$0x28203258
	pushl	$0x44363834
	jmp	.iquit

.i486sl:
	pushw	$0x004C
	pushl	$0x53363834
	jmp	.iquit

.i486sx2:
	pushl	$0x0075382E
	pushl	$0x30203258
	pushl	$0x53363834
	jmp	.iquit

.i486dx2wb:
	pushw	$0x0000
	pushl	$0x75382E30
	pushl	$0x20296B63
	pushl	$0x61622D65
	pushl	$0x74697277
	pushl	$0x28203258
	pushl	$0x44363834
	jmp	.iquit

.i486dx4wt:
	pushl	$0x0075362E
	pushl	$0x30202968
	pushl	$0x67756F72
	pushl	$0x68742D65
	pushl	$0x74697277
	pushl	$0x28203458
	pushl	$0x44363834
	jmp	.iquit

.i486dx4wb:
	pushw	$0x0000
	pushl	$0x75362E30
	pushl	$0x20296B63
	pushl	$0x61622D65
	pushl	$0x74697277
	pushl	$0x28203458
	pushl	$0x44363834
	jmp	.iquit

.i586:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.ip5pr
	cmpb	$0x01, %al
	je	.ip5
	cmpb	$0x02, %al
	je	.ip54c
	cmpb	$0x03, %al
	je	.ip24t
	cmpb	$0x04, %al
	je	.ip55c
	cmpb	$0x07, %al
	je	.ip54cs
	cmpb	$0x08, %al
	je	.itill
	pushl	$0x005D7373
	pushl	$0x616C6320
	pushl	$0x6D756974
	pushl	$0x6E65505B
	jmp	.iquit

.ip5pr:
	pushl	$0x00296570
	pushl	$0x79746F74
	pushl	$0x6F727028
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip5:
	pushw	$0x0075
	pushl	$0x382E3020
	pushl	$0x29355028
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip54c:
	pushl	$0x0075362E
	pushl	$0x30202943
	pushl	$0x34355028
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip24t:
	pushw	$0x0000
	pushl	$0x29543432
	pushl	$0x50282065
	pushl	$0x76697244
	pushl	$0x7265764F
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip55c:
	pushw	$0x0000
	pushl	$0x7538322E
	pushl	$0x30202943
	pushl	$0x35355028
	pushl	$0x20584D4D
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip54cs:
	pushw	$0x0075
	pushl	$0x35332E30
	pushl	$0x20295343
	pushl	$0x34355028
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.itill:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20296B6F
	pushl	$0x6F6D616C
	pushl	$0x6C695428
	pushl	$0x20584D4D
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.i686:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.ip6pr
	cmpb	$0x01, %al
	je	.ip6
	cmpb	$0x03, %al
	je	.iklmt
	cmpb	$0x05, %al
	je	.idess
	cmpb	$0x06, %al
	je	.idixn
	cmpb	$0x07, %al
	je	.ikatm
	cmpb	$0x08, %al
	je	.icopp
	cmpb	$0x09, %al
	je	.ipm13
	cmpb	$0x0A, %al
	je	.icasc
	cmpb	$0x0B, %al
	je	.itual
	pushl	$0x005D7373
	pushl	$0x616C6320
	pushl	$0x6F725020
	pushl	$0x6D756974
	pushl	$0x6E65505B
	jmp	.iquit

.ip6pr:
	pushl	$0x00296570
	pushl	$0x79746F74
	pushl	$0x6F727028
	pushl	$0x206F7250
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ip6:
	movl	%ebx, %eax
	shrl	$8, %eax
	andb	$0x0F, %al
	cmpb	$0x02, %al
	ja	.ip6a
/* steppings 1 and 2 */
	pushw	$0x0075
	pushl	$0x362E3020
	pushl	$0x29365028
	pushl	$0x206F7250
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.ip6a:
/* steppings 6, 7, 9 */
	pushl	$0x00007535
	pushl	$0x332E3020
	pushl	$0x29365028
	pushl	$0x206F7250
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.iklmt:
/* test for bit 13, though the stepping check should also succeed */
	movl	$1, %eax
	cpuid
	testw	$0x1000, %ax
	jnz	.ip6t
/* steppings 3 and 4 */
	pushl	$0x00007538
	pushl	$0x322E3020
	pushl	$0x29687461
	pushl	$0x6D616C4B
	pushl	$0x28204949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.ip6t:
/* stepping 2 */
	pushw	$0x0000
	pushl	$0x7535322E
	pushl	$0x30202954
	pushl	$0x36502820
	pushl	$0x65766972
	pushl	$0x44726576
	pushl	$0x4F204949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.idess:
/* test the S-cache size */
	movl	$2, %eax
	cpuid
	andb	$0xFF, %dl
	cmpb	$0x43, %dl
	jb	.icovg
	ja	.idrak
/* regular: steppings 0, 1, 2, 3 */
/* mobile: steppings 0 and 2 */
	pushw	$0x0000
	pushl	$0x7535322E
	pushl	$0x30202973
	pushl	$0x65747568
	pushl	$0x63736544
	pushl	$0x28204949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.idrak:
/* steppings 2 and 3 */
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x2029656B
	pushl	$0x61724428
	pushl	$0x206E6F65
	pushl	$0x58204949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.icovg:
/* steppings 0 and 1 */
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20296E6F
	pushl	$0x74676E69
	pushl	$0x766F4328
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.idixn:
	movl	%ebx, %eax
	shrl	$8, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.imend
	cmpb	$0x05, %al
	je	.imend
/* stepping A */
/* test the S-cache size */
	movl	$2, %eax
	cpuid
	andb	$0xFF, %dl
	cmpb	$0x42, %dl
	jb	.idxcl
/* 400MHz BGA and mPGA were 0.18u, others -- 0.25u */
	pushl	$0x00007538
	pushl	$0x312E302D
	pushl	$0x7535322E
	pushl	$0x3020296E
	pushl	$0x6F786944
	pushl	$0x28204949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.idxcl:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20296E6F
	pushl	$0x78694428
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit
.imend:
/* stepping 0 (SEPP) and 5 (PGA) */
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20296F6E
	pushl	$0x69636F64
	pushl	$0x6E654D28
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.ikatm:
/* test the S-cache size */
	movl	$2, %eax
	cpuid
	andb	$0xFF, %dl
	cmpb	$0x43, %dl
	ja	.itann
/* steppings 2 and 3 */
	pushl	$0x00007535
	pushl	$0x322E3020
	pushl	$0x2969616D
	pushl	$0x74614B28
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.itann:
/* steppings 2 and 3 */
	pushl	$0x00753532
	pushl	$0x2E302029
	pushl	$0x72656E6E
	pushl	$0x61542820
	pushl	$0x6E6F6558
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.icopp:
/* check the brand ID:
 * Celeron -- 0x1
 * Pentium III -- 0x2
 * Pentium III Xeon -- 0x3 */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x02, %bl
	ja	.icpxn
	jb	.icpcl
/* steppings 1, 3, 6, A */
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x29656E69
	pushl	$0x6D726570
	pushl	$0x706F4328
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.icpxn:
/* steppings 1, 3, 6 */
	pushl	$0x00753831
	pushl	$0x2E302029
	pushl	$0x656E696D
	pushl	$0x72657070
	pushl	$0x6F432820
	pushl	$0x6E6F6558
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.icpcl:
/* steppings 3, 6, A */
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x29656E69
	pushl	$0x6D726570
	pushl	$0x706F4328
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.ipm13:
/* check the brand ID:
 * Celeron M -- 0x12
 * Pentium M -- 0x16 */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x12, %bl
	je	.icm13
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x30202973
	pushl	$0x61696E61
	pushl	$0x4228204D
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.icm13:
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x30202973
	pushl	$0x61696E61
	pushl	$0x4228204D
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.icasc:
/* steppings 0, 1, 4 */
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x20297365
	pushl	$0x64616373
	pushl	$0x61432820
	pushl	$0x6E6F6558
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.itual:
/* check the brand ID:
 * Pentium III -- 0x4 (regular) or 0x6 (mobile)
 * Celeron -- 0x3 (stepping 1) or 0x7 (mobile) */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x03, %bl
	je	.itlcl
	cmpb	$0x07, %bl
	je	.itlcl
/* steppings 1 and 4 */
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x3020296E
	pushl	$0x6974616C
	pushl	$0x61755428
	pushl	$0x20494949
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.itlcl:
/* steppings 1 and 4 */
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x3020296E
	pushl	$0x6974616C
	pushl	$0x61755428
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.ipm09:
/* check the brand ID:
 * Celeron M -- 0x12
 * Pentium M -- 0x16 */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x12, %bl
	je	.icm09
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x3020296E
	pushl	$0x6168746F
	pushl	$0x4428204D
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.icm09:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x3020296E
	pushl	$0x6168746F
	pushl	$0x4428204D
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.i786:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.iwill
	cmpb	$0x01, %al
	je	.iwill
	cmpb	$0x02, %al
	je	.inort
	cmpb	$0x03, %al
	je	.ipres
	cmpb	$0x04, %al
	je	.iprsm
	pushw	$0x005D
	pushl	$0x7373616C
	pushl	$0x63203420
	pushl	$0x6D756974
	pushl	$0x6E65505B
	jmp	.iquit

.iwill:
/* Willamette family, check the brand ID:
 * Pentium 4 -- 0x08
 * Celeron -- 0x0A
 * Xeon (Foster) -- 0x0E
 * Xeon MP (Foster MP) -- 0x0B */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x08, %bl
	je	.iwlp4
	cmpb	$0x0A, %bl
	je	.iwlcl
	cmpb	$0x0E, %bl
	je	.iwlxn
	cmpb	$0x0B, %bl
	je	.iwlxnmp
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x205D796C
	pushl	$0x696D6166
	pushl	$0x20657474
	pushl	$0x656D616C
	pushl	$0x6C69575B
	jmp	.iquit

.iwlp4:
	pushw	$0x0000
	pushl	$0x7538312E
	pushl	$0x30202965
	pushl	$0x7474656D
	pushl	$0x616C6C69
	pushl	$0x57282034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.iwlcl:
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x29657474
	pushl	$0x656D616C
	pushl	$0x6C695728
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.iwlxn:
	pushl	$0x00753831
	pushl	$0x2E302029
	pushl	$0x72657473
	pushl	$0x6F462820
	pushl	$0x6E6F6558
	jmp	.iquit

.iwlxnmp:
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x2029504D
	pushl	$0x20726574
	pushl	$0x736F4628
	pushl	$0x20504D20
	pushl	$0x6E6F6558
	jmp	.iquit

.inort:
/* Northwood family, check the brand ID:
 * Pentium 4 -- 0x09 (regular) or 0x0E (mobile)
 * Celeron -- 0x08 (regular) or 0x0F (mobile)
 * Xeon (Prestonia) -- 0x0B
 * Xeon MP (Gallatin) -- 0x0C
 * NOTE: P4 EE (Gallatin) and regular P4 carry the same brand ID */
	movl	$1, %eax
	cpuid
	andb	$0xFF, %bl
	cmpb	$0x09, %bl
	je	.inrp4
	cmpb	$0x0E, %bl
	je	.inrp4
	cmpb	$0x08, %bl
	je	.inrcl
	cmpb	$0x0F, %bl
	je	.inrcl
	cmpb	$0x0B, %bl
	je	.inrxn
	cmpb	$0x0C, %bl
	je	.inrxnmp
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x30205D79
	pushl	$0x6C696D61
	pushl	$0x6620646F
	pushl	$0x6F776874
	pushl	$0x726F4E5B
	jmp	.iquit

.inrp4:
/* Northwood has no T-cache unlike Gallatin */
	xorl	%ebx, %ebx
	pushw	$0x0040
	call	iparser
	cmpl	%ebx, %eax
	je	.inrp4ee
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x646F6F77
	pushl	$0x6874726F
	pushl	$0x4E282034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit
.inrp4ee:
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x20296E69
	pushl	$0x74616C6C
	pushl	$0x61472820
	pushl	$0x45452034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.inrcl:
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x2029646F
	pushl	$0x6F776874
	pushl	$0x726F4E28
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.inrxn:
	pushl	$0x00007533
	pushl	$0x312E3020
	pushl	$0x2961696E
	pushl	$0x6F747365
	pushl	$0x72502820
	pushl	$0x6E6F6558
	jmp	.iquit

.inrxnmp:
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x3020296E
	pushl	$0x6974616C
	pushl	$0x6C614728
	pushl	$0x20504D20
	pushl	$0x6E6F6558
	jmp	.iquit

.ipres:
/* Prescott (initial) family, various checks need to be performed;
 * brand IDs are no longer supported by Intel for some arcane reason */
/* Celeron has 256Kb of S-cache */
	xorl	%ebx, %ebx
	pushw	$0x007A
	call	iparser
	cmpl	%ebx, %eax
	jne	.iprcl
/* P4 has no support for EM64T (extended feature flags, bit 29 of %edx) */
	movl	$0x80000001, %eax
	cpuid
	testl	$0x20000000, %edx
	jz	.iprp4
.iprxn:
/* must be Xeon (Nocona) */
	pushl	$0x00753930
	pushl	$0x2E302029
	pushl	$0x616E6F63
	pushl	$0x6F4E2820
	pushl	$0x6E6F6558
	jmp	.iquit

.iprcl:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x30202974
	pushl	$0x746F6373
	pushl	$0x65725028
	pushl	$0x206E6F72
	pushl	$0x656C6543
	jmp	.iquit

.iprp4:
	pushl	$0x00007539
	pushl	$0x302E3020
	pushl	$0x2974746F
	pushl	$0x63736572
	pushl	$0x50282034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.iprsm:
/* Prescott (continued), Prescott 2M and Smithfield families;
 * there is a real hardware mess, not all CPUs can be detected for sure */
/* check the stepping first */
	movl	%ebx, %eax
	andb	$0x0F, %al
	cmpb	$0x01, %al
	je	.ipr1m
	cmpb	$0x03, %al
	je	.ipr2m
	cmpb	$0x04, %al
	je	.ism
	cmpb	$0x09, %al
	je	.iprxn
	cmpb	$0x0A, %al
	je	.ipr2mxn
	pushw	$0x005D
	pushl	$0x7373616C
	pushl	$0x63203420
	pushl	$0x6D756974
	pushl	$0x6E65505B
	jmp	.iquit

.ipr1m:
/* Celeron still has 256Kb of S-cache */
	xorl	%ebx, %ebx
	pushw	$0x007A
	call	iparser
	cmpl	%ebx, %eax
	jne	.iprcl
/* Xeon MP (Potomac) supports T-cache unlike any other */
	pushw	$0x0040
	call	iparser
	cmpl	%ecx, %eax
	je	.iprxnmp
/* no luck here: treat as a P4, though could also be a Xeon (Nocona) or a
 * Xeon MP (Cranford) */
 	jmp	.iprp4

.iprxnmp:
	pushl	$0x00753930
	pushl	$0x2E302029
	pushl	$0x63616D6F
	pushl	$0x746F5028
	pushl	$0x20504D20
	pushl	$0x6E6F6558
	jmp	.iquit

.ipr2m:
/* again, no luck: treat as a P4, though could also be a Xeon (Irwindale) */
	pushw	$0x0075
	pushl	$0x39302E30
	pushl	$0x20294D32
	pushl	$0x2074746F
	pushl	$0x63736572
	pushl	$0x50282034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ism:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x30202964
	pushl	$0x6C656966
	pushl	$0x6874696D
	pushl	$0x53282034
	pushl	$0x206D7569
	pushl	$0x746E6550
	jmp	.iquit

.ipr2mxn:
	pushl	$0x00007539
	pushl	$0x302E3020
	pushl	$0x29656C61
	pushl	$0x646E6977
	pushl	$0x72492820
	pushl	$0x6E6F6558

.iquit:
	pushl	$0x206C6574
	pushw	$0x6E49
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.noitsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.noitsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.intelinfo
.noitsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.intelinfo

/* AMD CPUs */
.amd:
	movl	%ebx, %eax
	andb	$0x0F, %al
	cmpb	$0x04, %al
	je	.amd486
	cmpb	$0x05, %al
	je	.amd586
	cmpb	$0x06, %al
	je	.amd686
	cmpb	$0x0F, %al
	je	.amd786
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.amdquit

.amd486:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x03, %al
	je	.amd486dx2wt
	cmpb	$0x07, %al
	je	.amd486dx2wb
	cmpb	$0x08, %al
	je	.amd486dx4wt
	cmpb	$0x09, %al
	je	.amd486dx4wb
	cmpb	$0x0E, %al
	je	.amd5x86wt
	cmpb	$0x0F, %al
	je	.amd5x86wb
	pushl	$0x005D7373
	pushl	$0x616C6320
	pushl	$0x3638345B
	jmp	.amdquit

.amd486dx2wt:
	pushl	$0x0075372E
	pushl	$0x30202968
	pushl	$0x67756F72
	pushl	$0x68742D65
	pushl	$0x74697277
	pushl	$0x28203258
	pushl	$0x44363834
	jmp	.amdquit

.amd486dx2wb:
	pushw	$0x0000
	pushl	$0x75372E30
	pushl	$0x20296B63
	pushl	$0x61622D65
	pushl	$0x74697277
	pushl	$0x28203258
	pushl	$0x44363834
	jmp	.amdquit

.amd486dx4wt:
	pushl	$0x0075352E
	pushl	$0x30202968
	pushl	$0x67756F72
	pushl	$0x68742D65
	pushl	$0x74697277
	pushl	$0x28203458
	pushl	$0x44363834
	jmp	.amdquit

.amd486dx4wb:
	pushw	$0x0000
	pushl	$0x75352E30
	pushl	$0x20296B63
	pushl	$0x61622D65
	pushl	$0x74697277
	pushl	$0x28203458
	pushl	$0x44363834
	jmp	.amdquit

.amd5x86wt:
	pushl	$0x00007535
	pushl	$0x332E3020
	pushl	$0x29686775
	pushl	$0x6F726874
	pushl	$0x2D657469
	pushl	$0x72772820	
	pushl	$0x36387835
	jmp	.amdquit

.amd5x86wb:
	pushl	$0x00753533
	pushl	$0x2E302029
	pushl	$0x6B636162
	pushl	$0x2D657469
	pushl	$0x72772820
	pushl	$0x36387835
	jmp	.amdquit

.amd586:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x00, %al
	je	.amdssa5
	cmpb	$0x01, %al
	je	.amd5k86
	cmpb	$0x02, %al
	je	.amd5k86
	cmpb	$0x03, %al
	je	.amd5k86
	cmpb	$0x06, %al
	je	.amdk6
	cmpb	$0x07, %al
	je	.amdk6a
	cmpb	$0x08, %al
	je	.amdk62
	cmpb	$0x09, %al
	je	.amdk63
	cmpb	$0x0D, %al
	je	.amdk6p
	pushw	$0x0000
	pushl	$0x5D737361
	pushl	$0x6C632036
	pushl	$0x4B20726F
	pushl	$0x20354B5B
	jmp	.amdquit

.amdssa5:
	pushw	$0x0000
	pushl	$0x7535332E
	pushl	$0x30202935
	pushl	$0x2F415353
	pushl	$0x2820354B
	jmp	.amdquit

.amd5k86:
	pushl	$0x00753533
	pushl	$0x2E302029
	pushl	$0x36386B35
	pushl	$0x2820354B
	jmp	.amdquit

.amdk6:
	pushw	$0x0000
	pushl	$0x7535332E
	pushl	$0x3020364B
	jmp	.amdquit

.amdk6a:
	pushw	$0x0000
	pushl	$0x7535322E
	pushl	$0x3020364B
	jmp	.amdquit

.amdk62:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20297372
	pushl	$0x65626D61
	pushl	$0x68432820
	pushl	$0x322D364B
	jmp	.amdquit

.amdk63:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x20296874
	pushl	$0x6F6F7470
	pushl	$0x72616853
	pushl	$0x28204949
	pushl	$0x492D364B
	jmp	.amdquit

.amdk6p:
	movl	$0x80000006, %eax
	cpuid
	shrl	$16, %ecx
	cmpw	$256, %cx
	je	.amdk63p
	pushl	$0x00753831
	pushl	$0x2E30202B
	pushl	$0x322D364B
	jmp	.amdquit
.amdk63p:
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x202B4949
	pushl	$0x492D364B
	jmp	.amdquit

.amd686:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x01, %al
	je	.amdk7
	cmpb	$0x02, %al
	je	.amdk75
	cmpb	$0x03, %al
	je	.amdsptf
	cmpb	$0x04, %al
	je	.amdthdb
	cmpb	$0x06, %al
	je	.amdplmn
	cmpb	$0x07, %al
	je	.amdmrgn
	cmpb	$0x08, %al
	je	.amdthrb
	cmpb	$0x0A, %al
	je	.amdbrtn
	pushl	$0x00005D73
	pushl	$0x73616C63
	pushl	$0x206E6F6C
	pushl	$0x6874415B
	jmp	.amdquit

.amdk7:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x2029374B
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdk75:
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x2935374B
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdsptf:
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x29657269
	pushl	$0x66746970
	pushl	$0x5328206E
	pushl	$0x6F727544
	jmp	.amdquit

.amdthdb:
	pushl	$0x00007538
	pushl	$0x312E3020
	pushl	$0x29647269
	pushl	$0x62726564
	pushl	$0x6E756854
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdplmn:
	movl	$0x80000006, %eax
	cpuid
	shrl	$16, %ecx
	cmpw	$64, %cx
	jne	.amdmrgn
	pushl	$0x00753831
	pushl	$0x2E302029
	pushl	$0x6F6E696D
	pushl	$0x6F6C6150
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdmrgn:
	pushw	$0x0000
	pushl	$0x7538312E
	pushl	$0x3020296E
	pushl	$0x6167726F
	pushl	$0x4D28206E
	pushl	$0x6F727544
	jmp	.amdquit

.amdthrb:
	movl	$0x80000006, %eax
	cpuid
	shrl	$16, %ecx
	cmpw	$64, %cx
	je	.amdappl
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x64657262
	pushl	$0x6867756F
	pushl	$0x726F6854
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit
.amdappl:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x64657262
	pushl	$0x656C7070
	pushl	$0x4128206E
	pushl	$0x6F727544
	jmp	.amdquit

.amdbrtn:
	movl	$0x80000006, %eax
	cpuid
	shrl	$16, %ecx
	cmpw	$256, %cx
	je	.amdthrt
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x20296E6F
	pushl	$0x74726142
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit
.amdthrt:
	pushl	$0x00007533
	pushl	$0x312E3020
	pushl	$0x296E6F74
	pushl	$0x726F6854
	pushl	$0x28206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amd786:
	xorw	%si, %si
/* searching for a 8-bit brand ID (standard function 1);
 * 3 most-significant bits identify a CPU model, 5 least-significant bits
 * represent certain performance-related information */
	movl	$1, %eax
	cpuid
	movb	%bl, %cl
	andb	$0x1F, %cl
	andb	$0xE0, %bl
	movb	%cl, %bh
	cmpw	$0x0000, %bx
	cmovaw	%bx, %si
/* searching for a 12-bit brand ID (extended function 1);
 * 6 most-significant bits identify a CPU model, 6 least-significant bits
 * represent certain performance-related information */
	movl	$0x80000001, %eax
	cpuid
	movb	%bl, %cl
	andb	$0x3F, %cl
	shrw	$6, %bx
	andb	$0x3F, %bl
	movb	%cl, %bh
	cmpw	$0x0000, %bx
	cmovaw	%bx, %si
/* calculate the effective model */
	andb	$0xF0, %al
	shrb	$4, %al
	movb	%al, %cl
	shrl	$12, %eax
	andb	$0xF0, %al
	orb	%cl, %al
/* identify through it */
	cmpb	$0x04, %al
	je	.amdcl
	cmpb	$0x05, %al
	je	.amdsl
	cmpb	$0x07, %al
	je	.amdsl
	cmpb	$0x0C, %al
	je	.amdpr
	cmpb	$0x0F, %al
	je	.amdnw
	cmpb	$0x1C, %al
	je	.amdpl
	cmpb	$0x1F, %al
	je	.amdwn
	cmpb	$0x21, %al
	je	.amdeg
	cmpb	$0x23, %al
	je	.amdtl
	cmpb	$0x27, %al
	je	.amdsd
	cmpb	$0x2B, %al
	je	.amdmn
	cmpb	$0x2C, %al
	je	.amdpl
	cmpb	$0x2F, %al
	je	.amdvn
	pushw	$0x005D
	pushl	$0x7373616C
	pushl	$0x63203436
	pushl	$0x206E6F6C
	pushl	$0x6874415B
	jmp	.amdquit

.amdcl:
	pushw	$0x0000
	pushl	$0x7533312E
	pushl	$0x30202972
	pushl	$0x656D6D61
	pushl	$0x6877616C
	pushl	$0x43282034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdsl:
/* Sledgehammer family, check the brand ID:
 * Opteron UP -- 0x03
 * Opteron DP -- 0x04
 * Opteron MP -- 0x05
 * Athlon 64 FX -- 0x09 */
 	cmpw	$0x0003, %si
	je	.amdslup
	cmpw	$0x0004, %si
	je	.amdsldp
	cmpw	$0x0005, %si
	je	.amdslmp
	cmpw	$0x0009, %si
	je	.amdslfx
	pushl	$0x00753331
	pushl	$0x2E30205D
	pushl	$0x796C696D
	pushl	$0x61662072
	pushl	$0x656D6D61
	pushl	$0x68656764
	pushl	$0x656C535B
	jmp	.amdquit

.amdslup:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x72656D6D
	pushl	$0x61686567
	pushl	$0x64656C53
	pushl	$0x28205055
	pushl	$0x206E6F72
	pushl	$0x6574704F
	jmp	.amdquit

.amdsldp:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x72656D6D
	pushl	$0x61686567
	pushl	$0x64656C53
	pushl	$0x28205044
	pushl	$0x206E6F72
	pushl	$0x6574704F
	jmp	.amdquit

.amdslmp:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x72656D6D
	pushl	$0x61686567
	pushl	$0x64656C53
	pushl	$0x2820504D
	pushl	$0x206E6F72
	pushl	$0x6574704F
	jmp	.amdquit

.amdslfx:
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x20297265
	pushl	$0x6D6D6168
	pushl	$0x65676465
	pushl	$0x6C532820
	pushl	$0x58462034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdpr:
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x20297369
	pushl	$0x72615028
	pushl	$0x206E6F72
	pushl	$0x706D6553
	jmp	.amdquit

.amdnw:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x656C7473
	pushl	$0x61637765
	pushl	$0x4E282034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdpl:
	pushl	$0x00753930
	pushl	$0x2E302029
	pushl	$0x6F6D7265
	pushl	$0x6C615028
	movl	$0x80000001, %eax
	cpuid
	testl	$0x20000000, %edx
	jz	.amdpl32
	pushl	$0x20343600
	incl	%esp
.amdpl32:
	pushl	$0x206E6F72
	pushl	$0x706D6553
	jmp	.amdquit

.amdwn:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x30202972
	pushl	$0x65747365
	pushl	$0x68636E69
	pushl	$0x57282034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdeg:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x30203258
	pushl	$0x206E6F72
	pushl	$0x6574704F
	jmp	.amdquit

.amdtl:
	pushl	$0x00753930
	pushl	$0x2E302029
	pushl	$0x6F64656C
	pushl	$0x6F542820
	pushl	$0x32582034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdsd:
	pushl	$0x00007539
	pushl	$0x302E3020
	pushl	$0x296F6765
	pushl	$0x6944206E
	pushl	$0x61532820
	pushl	$0x58462034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdmn:
	pushl	$0x00753930
	pushl	$0x2E302029
	pushl	$0x72657473
	pushl	$0x6568636E
	pushl	$0x614D2820
	pushl	$0x32582034
	pushl	$0x36206E6F
	pushl	$0x6C687441
	jmp	.amdquit

.amdvn:
	pushw	$0x0000
	pushl	$0x7539302E
	pushl	$0x30202965
	pushl	$0x63696E65
	pushl	$0x56282034
	pushl	$0x36206E6F
	pushl	$0x6C687441

.amdquit:
	pushl	$0x20444D41
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.noamdtsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.noamdtsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo
.noamdtsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo

/* Cyrix CPUs;
 * 1) most CPUs don't support CPUID at all, and even those which do may have it
 *    disabled after reset (bit 7 of %ccr4 set to 0);
 * 2) on the other hand, all CPUs support vendor-specific registers, DIR0 and
 *    DIR1, which should be used instead of CPUID;
 * 3) CPUs, though designed by Cyrix, were manufactured by IBM, NSC, Texas
 *    Instruments and SGS-Thomson using different technical processes, probably
 *    detectable via DIR1 but that information is missing;
 * 4) only GXm and GXlv support standard function 2 and extended functions 0 to
 *    5, as well as their derivatives (NSC Geode GX1 and GX2);
 * 5) some VIA C3 CPUs promoted as "Cyrix III" and carried "CyrixInstead"
 *    vendor's ID, though in fact they were derivatives of the IDT C6 line and
 *    should be located appropriately. */
.cyrix:
/* use DIRs rather than CPUID in privileged mode */
	testl	$0x80000000, %edi
	jnz	.cxdirget
/* redirect to DIR-related data */
	movl	%ebx, %eax
	cmpb	$0x44, %al
	je	.cxmgx
	cmpb	$0x94, %al
	je	.cxm1sc
	cmpb	$0x05, %al
	je	.cxm1
	cmpb	$0x25, %al
	je	.cxm1l
	cmpb	$0x45, %al
	je	.cxgxm
	cmpb	$0x06, %al
	je	.cxm2
	cmpb	$0x66, %al
	je	.cntsam
	cmpb	$0x76, %al
	je	.cntsam2
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.cxquit

/* Centaur CPUs;
 * NOTE: Centaur was a subsidiary of IDT, and after that of VIA, so its CPUs
 *       were promoted under their brands, though carried the same vendor ID */
.centaur:
	movl	%ebx, %eax
	andb	$0x0F, %al
	cmpb	$0x05, %al
	je	.cnt586
	cmpb	$0x06, %al
	je	.cnt686
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.viaquit

.cnt586:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x04, %al
	je	.cntc6
	cmpb	$0x08, %al
	je	.cntc6p
	cmpb	$0x09, %al
	je	.cntc6pp
	pushw	$0x005D
	pushl	$0x36383520
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.idtquit

.cntc6:
	pushl	$0x00007535
	pushl	$0x332E3020
	pushl	$0x29364328
	pushl	$0x20706968
	pushl	$0x436E6957
	jmp	.idtquit

.cntc6p:
/* was either 0.35u or 0.25u */
	pushl	$0x00753532
	pushl	$0x2E302D75
	pushl	$0x35332E30
	pushl	$0x20292B36
	pushl	$0x43282032
	pushl	$0x20706968
	pushl	$0x436E6957
	jmp	.idtquit

.cntc6pp:
	pushl	$0x00753532
	pushl	$0x2E302033
	pushl	$0x20706968
	pushl	$0x436E6957
	jmp	.idtquit

.cnt686:
	movl	%ebx, %eax
	shrl	$4, %eax
	andb	$0x0F, %al
	cmpb	$0x06, %al
	jmp	.cntsam
	cmpb	$0x07, %al
	je	.cntsam2
	cmpb	$0x08, %al
	je	.cntezrt
	cmpb	$0x09, %al
	je	.cntnehm
	pushw	$0x005D
	pushl	$0x36383620
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.viaquit

.cntsam:
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x20296C65
	pushl	$0x756D6153
	pushl	$0x28203343
	jmp	.viaquit

.cntsam2:
/* Samuel has steppings < 8, while Ezra >= 8
 * (from "VIA C3 Ezra Processor Datasheet") */
	movl	%ebx, %eax
	shrl	$8, %eax
	cmpb	$8, %al
	jae	.cntezr
	pushl	$0x00753531
	pushl	$0x2E302029
	pushl	$0x32206C65
	pushl	$0x756D6153
	pushl	$0x28203343
	jmp	.viaquit
.cntezr:
	pushl	$0x00753531
	pushl	$0x2E302029
	pushl	$0x61727A45
	pushl	$0x28203343
	jmp	.viaquit

.cntezrt:
	pushw	$0x0075
	pushl	$0x33312E30
	pushl	$0x2029542D
	pushl	$0x61727A45
	pushl	$0x28203343
	jmp	.viaquit

.cntnehm:
	pushl	$0x00753331
	pushl	$0x2E302029
	pushl	$0x6861696D
	pushl	$0x6568654E
	pushl	$0x28203343
	jmp	.viaquit

.idtquit:
	pushl	$0x20544449
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.noidttsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.noidttsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo
.noidttsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo

.viaquit:
	pushl	$0x20414956
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.noviatsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.noviatsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo
.noviatsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.amdinfo

/* Transmeta CPUs */
.transmeta:
	movl	%ebx, %eax
	cmpb	$0x45, %al
	je	.tmcrus
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.tmquit

.tmcrus:
	pushl	$0x0000656F
	pushl	$0x73757243

.tmquit:
	pushl	$0x20617465
	pushl	$0x6D736E61
	pushw	$0x7254
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.notmtsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.notmtsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.notmtsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* Rise CPUs;
 * (from "Rise iDragon mP6 Microprocessor Data Sheet") */
.rise:
	movl	%ebx, %eax
	cmpb	$0x05, %al
	je	.rsfrst
	cmpb	$0x25, %al
	je	.rskirn
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.rsquit

.rsfrst:
	pushw	$0x0075
	pushl	$0x35322E30
	pushl	$0x2036506D
	pushl	$0x206E6F67
	pushl	$0x61724469
	jmp	.rsquit

.rskirn:
	pushw	$0x0075
	pushl	$0x38312E30
	pushl	$0x20296E69
	pushl	$0x72694B28
	pushl	$0x2036506D
	pushl	$0x206E6F67
	pushl	$0x61724469

.rsquit:
	pushl	$0x20657369
	pushw	$0x5200
	incl	%esp
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.norstsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.norstsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.norstsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* NexGen CPUs;
 * NOTE: both 586 and 586FP have no hardware logic to support CPUID, but a
 *       BIOS may load an appropriate emulating microcode on start-up */
.nexgen:
/* nothing to do here, go to non-CPUID data */
	jmp	.nx586_print

/* UMC CPUs */
.umc:
	movl	%ebx, %eax
	cmpb	$0x14, %al
	je	.u5dx
	cmpb	$0x24, %al
	je	.u5sx
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.umcquit

.u5dx:
	pushw	$0x0075
	pushl	$0x362E3020
	pushl	$0x58443555
	jmp	.umcquit

.u5sx:
	pushw	$0x0075
	pushl	$0x362E3020
	pushl	$0x58533555

.umcquit:
	pushl	$0x20434D55
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode;
 * TSC is unsupported, but may calculate alternatively
 * (no datasheet, estimated empirically) */
	testl	$0x80000000, %edi
	jz	.noumccnt
	pushw	$4
	call	calcnt
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.noumccnt
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.noumccnt:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* NSC (National Semiconductor) CPUs;
 * NOTES:
 *   1) only GX1 and GX2 are "Geode by NSC", others are "CyrixInstead";
 *   2) in August of 2003, the whole processor department was sold to AMD. */
.nsc:
	movl	%ebx, %eax
	cmpb	$0x45, %al
	je	.nsgx1
	cmpb	$0x55, %al
	je	.nsgx2
	pushw	$0x005D
	pushl	$0x6E776F6E
	pushl	$0x6B6E755B
	jmp	.nscquit

.nsgx1:
	pushl	$0x00753831
	pushl	$0x2E302031
	pushl	$0x58472065
	pushl	$0x646F6547
	jmp	.nscquit

.nsgx2:
	pushl	$0x00753331
	pushl	$0x2E302032
	pushl	$0x58472065
	pushl	$0x646F6547

.nscquit:
	pushl	$0x2043534E
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.nonsctsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.nonsctsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.nonsctsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* SiS CPUs;
 * NOTE: this is a family of single-chip computer designs based upon the Rise
 *       mP6 core (licenced by SiS in October of 1999) */
.sis:
	pushl	$0x00003638
	pushl	$0x00783535
	pushl	$0x20536953
	pushl	%esp
	leal	-84(%ebp), %eax
	pushl	%eax
	call	printf
/* obtain and print processor's frequency if in privileged mode */
	testl	$0x80000000, %edi
	jz	.nosistsc
	call	caltsc
	xorl	%edx, %edx
	cmpl	%edx, %eax
	je	.nosistsc
	movl	$1000, %ecx
	divl	%ecx
	pushl	%edx
	pushl	%eax
	leal	-96(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat
.nosistsc:
	leal	-98(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* obtain and print Intel's extended information;
 * NOTE: implemented in Pentium Pro and up the way (from "Intel Processor
 *       Identification and the CPUID Instruction Application Note") */
.intelinfo:
	movl	$2, %eax
	cpuid
	movl	%eax, %esi
	andw	$0x00FF, %si
	cmpw	$0x0000, %si
	je	.cpufeat

/* I-cache (in uops) */
	xorl	%ebx, %ebx
	pushw	$0x0070
	call	iparser
	cmpl	%ebx, %eax
	je	.il1t1
	pushl	$0x00000008
	pushl	$0x0000000C
	jmp	.il1iuprint

.il1t1:
	pushw	$0x00071
	call	iparser
	cmpl	%ebx, %eax
	je	.il1t2
	pushl	$0x00000008
	pushl	$0x00000010
	jmp	.il1iuprint

.il1t2:
	pushw	$0x0072
	call	iparser
	cmpl	%ebx, %eax
	je	.il1icache
	pushl	$0x00000008
	pushl	$0x00000020
	jmp	.il1iuprint

.il1iuprint:
	leal	-124(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.intel_dcache

/* I-cache (in bytes) */
.il1icache:
	xorl	%ebx, %ebx
	pushw	$0x0006
	call	iparser
	cmpl	%ebx, %eax
	je	.il1i1
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000008
	jmp	.il1ibprint

.il1i1:
	pushw	$0x0008
	call	iparser
	cmpl	%ebx, %eax
	je	.il1i2
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000010
	jmp	.il1ibprint

.il1i2:
	pushw	$0x0030
	call	iparser
	cmpl	%ebx, %eax
	je	.il1i_unk
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000020
	jmp	.il1ibprint

.il1i_unk:
	pushl	$0x00000049
	leal	-190(%ebp), %eax
	call	printf
	jmp	.intel_dcache

.il1ibprint:
	pushl	$0x00000049
	leal	-168(%ebp), %eax
	pushl	%eax
	call	printf

/* D-cache */
.intel_dcache:
	xorl	%ebx, %ebx
	pushw	$0x000A
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d1
	pushl	$0x00000020
	pushl	$0x00000002
	pushl	$0x00000008
	jmp	.il1dprint

.il1d1:
	pushw	$0x000C
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d2
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000010
	jmp	.il1dprint

.il1d2:
	pushw	$0x002C
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d3
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000020
	jmp	.il1dprint

.il1d3:
	pushw	$0x0060
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d4
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000010
	jmp	.il1dprint

.il1d4:
	pushw	$0x0066
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d5
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000008
	jmp	.il1dprint

.il1d5:
	pushw	$0x0067
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d6
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000010
	jmp	.il1dprint

.il1d6:
	pushw	$0x0068
	call	iparser
	cmpl	%ebx, %eax
	je	.il1d_unk
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000020
	jmp	.il1dprint

.il1d_unk:
	pushl	$0x00000044
	leal	-190(%ebp),%eax
	pushl	%eax
	call	printf
	jmp	.il1itlb

.il1dprint:
	pushl	$0x00000044
	leal	-168(%ebp), %eax
	pushl	%eax
	call	printf

/* I-cache TLB */
.il1itlb:
	xorl	%ebx, %ebx
	pushw	$0x0001
	call	iparser
	cmpl	%ebx, %eax
	je	.il1it1
	pushl	$0x00000004
	pushl	$0x00000020
	jmp	.il1itprint

.il1it1:
	pushw	$0x0050
	call	iparser
	cmpl	%ebx, %eax
	je	.il1it2
	pushl	$0x00000040
	pushl	$0x00000040
	jmp	.il1itprint

.il1it2:
	pushw	$0x0051
	call	iparser
	cmpl	%ebx, %eax
	je	.il1it3
	pushl	$0x00000080
	pushl	$0x00000080
	jmp	.il1itprint

.il1it3:
	pushw	$0x0052
	call	iparser
	cmpl	%ebx, %eax
	je	.il1it4
	pushl	$0x00000100
	pushl	$0x00000100
	jmp	.il1itprint

.il1it4:
	pushw	$0x00B0
	call	iparser
	cmpl	%ebx, %eax
	je	.il1it_unk
	pushl	$0x00000004
	pushl	$0x00000080
	jmp	.il1itprint

.il1it_unk:
	pushl	$0x00000049
	leal	-328(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.il1dtlb

.il1itprint:
	pushl	$0x00000049
	leal	-302(%ebp), %eax
	pushl	%eax
	call	printf

/* D-cache TLB */
.il1dtlb:
	xorl	%ebx, %ebx
	pushw	$0x0003
	call	iparser
	cmpl	%ebx, %eax
	je	.il1dt1
	pushl	$0x00000004
	pushl	$0x00000040
	jmp	.il1dtprint

.il1dt1:
	pushw	$0x005B
	call	iparser
	cmpl	%ebx, %eax
	je	.il1dt2
	pushl	$0x00000040
	pushl	$0x00000040
	jmp	.il1dtprint

.il1dt2:
	pushw	$0x005C
	call	iparser
	cmpl	%ebx, %eax
	je	.il1dt3
	pushl	$0x00000080
	pushl	$0x00000080
	jmp	.il1dtprint

.il1dt3:
	pushw	$0x005D
	call	iparser
	cmpl	%ebx, %eax
	je	.il1dt4
	pushl	$0x00000100
	pushl	$0x00000100
	jmp	.il1dtprint

.il1dt4:
	pushw	$0x00B3
	call	iparser
	cmpl	%ebx, %eax
	je	.il1dt_unk
	pushl	$0x00000004
	pushl	$0x00000080
	jmp	.il1itprint

.il1dt_unk:
	pushl	$0x00000049
	leal	-328(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.il2cache

.il1dtprint:
	pushl	$0x00000044
	leal	-302(%ebp), %eax
	pushl	%eax
	call	printf

/* S-cache */
.il2cache:
	xorl	%ebx, %ebx
	pushw	$0x0039
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u1
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000080
	jmp	.il2uprint

.il2u1:
	pushw	$0x003B
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u2
	pushl	$0x00000040
	pushl	$0x00000002
	pushl	$0x00000080
	jmp	.il2uprint

.il2u2:
	pushw	$0x003C
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u3
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000100
	jmp	.il2uprint

.il2u3:
	pushw	$0x0041
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u4
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000080
	jmp	.il2uprint

.il2u4:
	pushw	$0x0042
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u5
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000100
	jmp	.il2uprint

.il2u5:
	pushw	$0x0043
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u6
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000200
	jmp	.il2uprint

.il2u6:
	pushw	$0x0044
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u7
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000400
	jmp	.il2uprint

.il2u7:
	pushw	$0x0045
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u8
	pushl	$0x00000020
	pushl	$0x00000004
	pushl	$0x00000800
	jmp	.il2uprint

.il2u8:
	pushw	$0x0078
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u9
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000400
	jmp	.il2uprint

.il2u9:
	pushw	$0x0079
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u10
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000080
	jmp	.il2uprint

.il2u10:
	pushw	$0x007A
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u11
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000100
	jmp	.il2uprint

.il2u11:
	pushw	$0x007B
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u12
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000200
	jmp	.il2uprint

.il2u12:
	pushw	$0x007C
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u13
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000400
	jmp	.il2uprint

.il2u13:
	pushw	$0x007D
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u14
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000800
	jmp	.il2uprint

.il2u14:
	pushw	$0x007F
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u15
	pushl	$0x00000040
	pushl	$0x00000002
	pushl	$0x00000200
	jmp	.il2uprint

.il2u15:
	pushw	$0x0082
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u16
	pushl	$0x00000020
	pushl	$0x00000008
	pushl	$0x00000100
	jmp	.il2uprint

.il2u16:
	pushw	$0x0083
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u17
	pushl	$0x00000020
	pushl	$0x00000008
	pushl	$0x00000200
	jmp	.il2uprint

.il2u17:
	pushw	$0x0084
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u18
	pushl	$0x00000020
	pushl	$0x00000008
	pushl	$0x00000400
	jmp	.il2uprint

.il2u18:
	pushw	$0x0085
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u19
	pushl	$0x00000020
	pushl	$0x00000008
	pushl	$0x00000800
	jmp	.il2uprint

.il2u19:
	pushw	$0x0086
	call	iparser
	cmpl	%ebx, %eax
	je	.il2u20
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000200
	jmp	.il2uprint

.il2u20:
	pushw	$0x0087
	call	iparser
	cmpl	%ebx, %eax
	je	.cpufeat
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000400

.il2uprint:
	pushl	$0x00000053
	leal	-168(%ebp), %eax
	pushl	%eax
	call	printf

/* T-cache */
.il3cache:
	xorl	%ebx, %ebx
	pushw	$0x0022
	call	iparser
	cmpl	%ebx, %eax
	je	.il3u1
	pushl	$0x00000040
	pushl	$0x00000004
	pushl	$0x00000200
	jmp	.il3uprint

.il3u1:
	pushw	$0x0023
	call	iparser
	cmpl	%ebx, %eax
	je	.il3u2
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000400
	jmp	.il3uprint

.il3u2:
	pushw	$0x0025
	call	iparser
	cmpl	%ebx, %eax
	je	.il3u3
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00000800
	jmp	.il3uprint

.il3u3:
	pushw	$0x0029
	call	iparser
	cmpl	%ebx, %eax
	je	.cpufeat
	pushl	$0x00000040
	pushl	$0x00000008
	pushl	$0x00001000

.il3uprint:
	pushl	$0x00000054
	leal	-168(%ebp), %eax
	pushl	%eax
	call	printf
	jmp	.cpufeat

/* obtain and print AMD's extended information;
 * implemented in K5 and up the way, also in IDT WinChips and VIA C3s
/* (from "AMD Processor Recognition Application Note") */
.amdinfo:
	movl	$0x80000000, %eax
	cpuid
	cmpl	$0x80000005, %eax
	jb	.cpufeat
/* I-cache */
	movl	$0x80000005, %eax
	cpuid
	movl	%edx, %esi
	andl	$0x000000FF, %esi
	pushl	%esi
	movl	%edx, %esi
	andl	$0x0000FF00, %esi
	shrl	$8, %esi
	pushl	%esi
	movl	%edx, %esi
	andl	$0x00FF0000, %esi
	shrl	$16, %esi
	pushl	%esi
	movl	%edx, %esi
	andl	$0xFF000000, %esi
	shrl	$24, %esi
	pushl	%esi
	pushl	$0x00000049
	leal	-252(%ebp), %eax
	pushl	%eax
	call	printf
/* D-cache */
	movl	$0x80000005, %eax
	cpuid
	movl	%ecx, %esi
	andl	$0x000000FF, %esi
	pushl	%esi
	movl	%ecx, %esi
	andl	$0x0000FF00, %esi
	shrl	$8, %esi
	pushl	%esi
	movl	%ecx, %esi
	andl	$0x00FF0000, %esi
	shrl	$16, %esi
	pushl	%esi
	movl	%ecx, %esi
	andl	$0xFF000000, %esi
	shrl	$24, %esi
	pushl	%esi
	pushl	$0x00000044
	leal	-252(%ebp), %eax
	pushl	%eax
	call	printf
/* TLB of I-cache and D-cache */
	movl	$0x80000005, %eax
	cpuid
	movl	%ebx, %esi
	andl	$0x0000FF00, %esi
	shrl	$8, %esi
	pushl	%esi
	movl	%ebx, %esi
	andl	$0x000000FF, %esi
	pushl	%esi
	cmpb	$0xFF, 4(%esp)
	jne	.aitlb_end
	movw	%si, 4(%esp)
.aitlb_end:
	pushl	$0x00000049
	leal	-302(%ebp), %eax
	pushl	%eax
	call	printf
	movl	$0x80000005, %eax
	cpuid
	movl	%ebx, %esi
	andl	$0xFF000000, %esi
	shrl	$24, %esi
	pushl	%esi
	movl	%ebx, %esi
	andl	$0x00FF0000, %esi
	shrl	$16, %esi
	pushl	%esi
	cmpb	$0xFF, 4(%esp)
	jne	.adtlb_end
	movw	%si, 4(%esp)
.adtlb_end:
	pushl	$0x00000044
	leal	-302(%ebp), %eax
	pushl	%eax
	call	printf
/* S-cache */
	movl	$0x80000000, %eax
	cpuid
	cmpl	$0x80000006, %eax
	jb	.cpufeat
	movl	$0x80000006, %eax
	cpuid
	movl	%ecx, %esi
	andl	$0x000000FF, %esi
	pushl	%esi
	movl	%ecx, %esi
	andl	$0x00000F00, %esi
	shrl	$8, %esi
	pushl	%esi
	movl	%ecx, %esi
	andl	$0x0000F000, %esi
	shrl	$12, %esi
	cmpw	$0x0006, %si
	jne	.l2w16
	movw	$0x0008, %si
	jmp	.l2wend
.l2w16:
	cmpw	$0x0008, %si
	jne	.l2wend
	movw	$0x0010, %si
.l2wend:
	pushl	%esi
	movl	%ecx, %esi
	andl	$0xFFFF0000, %esi
	shrl	$16, %esi
	pushl	%esi
	pushl	$0x00000053
	leal	-252(%ebp), %eax
	pushl	%eax
	call	printf
/* TLB of S-cache */
	movl	$0x80000006, %eax
	cpuid
	movl	%ebx, %esi
	andl	$0x0000F000, %esi
	shrl	$12, %esi
	pushl	%esi
	movl	%ebx, %esi
	andl	$0x00000FFF, %esi
	pushl	%esi
	cmpb	$0xFF, 4(%esp)
	jne	.astlb_end
	movw	%si, 4(%esp)
.astlb_end:
	pushl	$0x00000053
	leal	-302(%ebp), %eax
	pushl	%eax
	call	printf

/* print the features obtained before */
.cpufeat:
	pushw	$0x0000
	testw	$0x0100, %di
	jz	.featsse2
	pushl	$0x33455353
	pushw	$0x2000
	incl	%esp
.featsse2:
	testw	$0x0080, %di
	jz	.featsse
	pushl	$0x32455353
	pushw	$0x2000
	incl	%esp
.featsse:
	testw	$0x0040, %di
	jz	.featx3dn
	pushl	$0x45535320
.featx3dn:
	testw	$0x0020, %di
	jz	.feat3dn
	pushl	$0x2B21776F
	pushl	$0x4E443320
.feat3dn:
	testw	$0x0010, %di
	jz	.featxmmx
	pushl	$0x21776F4E
	pushl	$0x44332000
	incl	%esp
.featxmmx:
	testw	$0x0008, %di
	jz	.featmmx
	pushl	$0x2B584D4D
	pushw	$0x2000
	incl	%esp
.featmmx:
	testw	$0x0004, %di
	jz	.featcmov
	pushl	$0x584D4D20
.featcmov:
	testw	$0x0002, %di
	jz	.featfpu
	pushl	$0x564F4D43
	pushw	$0x2000
	incl	%esp
.featfpu:
	testw	$0x0001, %di
	jz	.featprint
	pushl	$0x55504620
.featprint:
	pushl	%esp
	leal	-342(%ebp), %eax
	pushl	%eax
	call	printf

.capquit:
	movl	%ebp, %esp
	andl	$0x3FFFFFFF, %edi
	movl	%edi, %eax
	popl	%edi
	popl	%esi
	popl	%ebp
	popl	%ebx
	ret

/* a parsing routine to retrieve extended information from Intel CPUs (using
 * standard function 2), needed because those descriptors have no fixed
 * locations (what a nonsense...) - RMH
 * input: byte-wide descriptor to search for (word, in stack)
 * output: descriptor if found, zero if not (double word, in %eax) */
.globl iparser
iparser:
	pushl	%ebx
	pushl	%ebp
	pushl	%esi
	pushl	%edi
	movw	20(%esp), %si
	movl	$2, %eax
	cpuid
	testl	$0x80000000, %eax
	jnz	.stage1
	shrl	$8, %eax
	movl	%eax, %ebp
	movl	%eax, %edi
	andw	$0x00FF, %ax
	shrl	$8, %ebp
	andw	$0x00FF, %bp
	shrl	$16, %edi
	andw	$0x00FF, %di
	cmpw	%si, %ax
	je	.found
	cmpw	%si, %bp
	je	.found
	cmpw	%si, %di
	je	.found
.stage1:
	testl	$0x80000000, %ebx
	jnz	.stage2
	movl	%ebx, %eax
	movl	%ebx, %ebp
	movl	%ebx, %edi
	andw	$0x00FF, %bx
	shrl	$8, %eax
	andw	$0x00FF, %ax
	shrl	$16, %ebp
	andw	$0x00FF, %bp
	shrl	$24, %edi
	andw	$0x00FF, %di
	cmpw	%si, %bx
	je	.found
	cmpw	%si, %ax
	je	.found
	cmpw	%si, %bp
	je	.found
	cmpw	%si, %di
	je	.found
.stage2:
	testl	$0x80000000, %ecx
	jnz	.stage3
	movl	%ecx, %eax
	movl	%ecx, %ebp
	movl	%ecx, %edi
	andw	$0x00FF, %cx
	shrl	$8, %eax
	andw	$0x00FF, %ax
	shrl	$16, %ebp
	andw	$0x00FF, %bp
	shrl	$24, %edi
	andw	$0x00FF, %di
	cmpw	%si, %cx
	je	.found
	cmpw	%si, %ax
	je	.found
	cmpw	%si, %bp
	je	.found
	cmpw	%si, %di
	je	.found
.stage3:
	testl	$0x80000000, %edx
	jnz	.notfound
	movl	%edx, %eax
	movl	%edx, %ebp
	movl	%edx, %edi
	andw	$0x00FF, %dx
	shrl	$8, %eax
	andw	$0x00FF, %ax
	shrl	$16, %ebp
	andw	$0x00FF, %bp
	shrl	$24, %edi
	andw	$0x00FF, %di
	cmpw	%si, %dx
	je	.found
	cmpw	%si, %ax
	je	.found
	cmpw	%si, %bp
	je	.found
	cmpw	%si, %di
	je	.found
.notfound:
	xorl	%eax, %eax
	jmp	.prsquit
.found:
	xorl	%eax, %eax
	movw	%si, %ax
.prsquit:
	popl	%edi
	popl	%esi
	popl	%ebp
	popl	%ebx
	ret

/* calculate processor's frequency using its Time-Stamp Counter and a standard
 * i8254-compatible timer (use counter 2)
 * input: none
 * output: processor's frequency in KHz (double word, in %eax)
 * (for further information regarding i8254 refer to "Intel 8254 Programmable
 * Interval Timer" datasheet; TSC is documented in various datasheets)
 * P.S. This is the fastest and most accurate calculator I've ever written,
 * with auto-calibration, and it doesn't use FPU at all! - RMH */
.globl caltsc
caltsc:
/* save the registers */
	pushl	%ebx
	pushl	%ebp
	pushl	%esi

/* verify support for CPUID */
	pushfl
	popl	%eax
	movl	%eax, %ebx
	xorl	$0x00200000, %eax
	pushl	%eax
	popfl
	pushfl
	popl	%eax
	cmpl	%eax, %ebx
	je	.tscnot
	pushl	%ebx
	popfl

/* verify support for TSC */
	movl	$1, %eax
	cpuid
	testw	$0x0010, %dx
	jz	.tscnot

/* filter out CPUs with broken TSCs */
	xorl	%eax, %eax
	cpuid
/* AMD K5 model 0 */
	cmpl	$0x444D4163, %ecx
	jne	.tscidt
	movl	$1, %eax
	cpuid
	andw	$0x0FF0, %ax
	cmpw	$0x0500, %ax
	jne	.tscgood
	jmp	.tscnot
.tscidt:
/* IDT WinChip */
	cmpl	$0x736C7561, %ecx
	jne	.tscgood
	movl	$1, %eax
	cpuid
	andw	$0x0FF0, %ax
	cmpw	$0x0540, %ax
	jne	.tscgood
	jmp	.tscnot

.tscgood:
/* set the initial loop count */
	movl	$0x00002000, %ebp
/* disable interrupts */
	cli
/* disable clock to the counter */
	inb	$0x61, %al
	andb	$0xFE, %al
	outb	%al, $0x61

.xloop:
/* advance the loop count */
	shll	$1, %ebp
	movl	%ebp, %ecx
/* set up the counter for a non-repeating countdown from 0xFFFF (mode 0,
 * non-BCD, LSB first, MSB second) */
	movb	$0xB0, %al
	outb	%al, $0x43
	movb	$0xFF, %al
	outb	%al, $0x42
	outb	%al, $0x42
/* enable clock to the counter */
	inb	$0x61, %al
	orb	$0x01, %al
	outb	%al, $0x61
/* obtain and save the initial TSC value;
 * don't care about the high double word, only a 157GHz or faster CPU may
 * spend 2^32 or more cycles during 0x7F00 ticks of i8254 */
	rdtsc
	movl	%eax, %esi
/* the delay loop;
 * integer signed divide is the most expensive [integer] instruction, also
 * it's non-pipelined or pipelined partially usually */
	xorl	%edx, %edx
	xorl	%eax, %eax
	movl	$1, %ebx
.yloop:
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	loopl	.yloop

/* disable clock to the counter */
	inb	$0x61, %al
	andb	$0xFE, %al
	outb	%al, $0x61
/* obtain and save the final TSC value */
	rdtsc
	movl	%eax, %edx
/* send the latch command to the counter */
	movb	$0x80, %al
	outb	%al, $0x43
/* read the count (LSB first, MSB second) */
	inb	$0x42, %al
	movb	%al, %bl
	inb	$0x42, %al
	movb	%al, %bh
/* calculate the clocks elapsed;
 * auto-calibration: if %bx < 0x7F00, increase the loop count by 2, and
 * rewind */
	notw	%bx
	cmpw	$0x7F00, %bx
	jb	.xloop

/* reset and stop the counter */
	movb	$0xB0, %al
	outb	%al, $0x43
	xorb	%al, %al
	outb	%al, $0x42
	outb	%al, $0x42
/* enable clock to the counter */
	inb	$0x61, %al
	orb	$0x01, %al
	outb	%al, $0x61
/* enable interrupts */
	sti
/* calculate the TSC delta */
	cmpl	%esi, %edx
	jbe	.tsccross
	subl	%esi, %edx
	jmp	.tsccalc

.tsccross:
	notl	%esi
	addl	%esi, %edx

.tsccalc:
/* calculate the frequency (in KHz, to avoid overflowing) */
	movl	%edx, %eax
	movl	$1193182, %ecx
	mull	%ecx
	imull	$1000, %ebx
	divl	%ebx
	jmp	.tscret

.tscnot:
/* not able to use TSC */
	xorl	%eax, %eax
.tscret:
/* restore the registers and return */
	popl	%esi
	popl	%ebp
	popl	%ebx
	ret

/* calculate processor's frequency using instruction clock count method and
 * a standard i8254-compatible timer (use counter 2)
 * input: processor's code (word, in stack)
 *   0x1 -- Cyrix 486 series;
 *   0x2 -- Cyrix 5x86, MediaGX, GXm, GXlv;
 *   0x3 -- Cyrix 6x86, 6x86L, 6x86MX, 6x86MII;
 *   0x4 -- UMC U5DX, U5SX;
 * output: processor's frequency in KHz (double word, in %eax) */
.globl calcnt
calcnt:
/* save the registers */
	pushl	%ebx
	pushl	%ebp
	pushl	%esi
/* take a processor's code from the stack */
	movw	20(%esp), %si
/* return if the code is < 0x1 or > 0x4 */
	xorl	%eax, %eax
	movw	$0x0001, %dx
	cmpw	%dx, %si
	jb	.cntret
	movw	$0x0004, %dx
	cmpw	%dx, %si
	ja	.cntret
/* set the initial loop count */
	movl	$0x00001000, %ebp
/* disable interrupts */
	cli
/* disable clock to the counter */
	inb	$0x61, %al
	andb	$0xFE, %al
	outb	%al, $0x61

.cxloop:
/* advance the loop count */
	shll	$1, %ebp
	movl	%ebp, %ecx
/* set up the counter for a non-repeating countdown from 0xFFFF (mode 0,
 * non-BCD, LSB first, MSB second) */
	movb	$0xB0, %al
	outb	%al, $0x43
	movb	$0xFF, %al
	outb	%al, $0x42
	outb	%al, $0x42
/* enable clock to the counter */
	inb	$0x61, %al
	orb	$0x01, %al
	outb	%al, $0x61
/* the master loop */
	xorl	%edx, %edx
	xorl	%eax, %eax
	movl	$1, %ebx
.cyloop:
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	idivl	%ebx
	loopl	.cyloop

/* disable clock to the counter */
	inb	$0x61, %al
	andb	$0xFE, %al
	outb	%al, $0x61
/* send the latch command to the counter */
	movb	$0x80, %al
	outb	%al, $0x43
/* read the count (LSB first, MSB second) */
	inb	$0x42, %al
	movb	%al, %bl
	inb	$0x42, %al
	movb	%al, %bh
/* calculate clocks elapsed;
 * auto-calibration: if %bx < 0x7F00, increase the loop count by 2, and
 * rewind */
	notw	%bx
	cmpw	$0x7F00, %bx
	jb	.cxloop

/* reset and stop the counter */
	movb	$0xB0, %al
	outb	%al, $0x43
	xorb	%al, %al
	outb	%al, $0x42
	outb	%al, $0x42
/* enable clock to the counter */
	inb	$0x61, %al
	orb	$0x01, %al
	outb	%al, $0x61
/* enable interrupts */
	sti
/* define the count multipliers;
 * (clocks per IDIV are in %ecx, per LOOP -- in %edx) */
	cmpw	$1, %si
	je	.calcx486
	cmpw	$2, %si
	je	.calcx586
	cmpw	$3, %si
	je	.calcx686
	cmpw	$4, %si
	je	.calu5
	xorl	%ecx, %ecx
	xorl	%edx, %edx
	jmp	.calnum

.calcx486:
	movl	$40, %ecx
	movl	$8, %edx
	jmp	.calnum

.calcx586:
	movl	$45, %ecx
	movl	$2, %edx
	jmp	.calnum

.calcx686:
	movl	$17, %ecx
	movl	$1, %edx
	jmp	.calnum

.calu5:
	movl	$7, %ecx
	movl	$4, %edx

.calnum:
/* calculate the number of clocks spent */
	imull	$5, %ebp, %eax
	imull	%ecx, %eax
	imull	%edx, %ebp
	addl	%ebp, %eax
/* calculate the frequency (in KHz, though overflowing isn't the issue) */
	movl	$1193182, %ecx
	mull	%ecx
	imull	$1000, %ebx
	divl	%ebx
.cntret:
/* restore the registers and return */
	popl	%esi
	popl	%ebp
	popl	%ebx
	ret
