/*
**  RAMspeed -- a cache & memory benchmarking tool
**
**  (for uniprocessor machines running UNIX-like operating systems)
**
**
**  This software is distributed under the terms of The Alasir Licence (TAL).
**  You should have received a copy of the licence together with the software.
**  If not, you should download it from http://www.alasir.com/licence/TAL.txt
**
**
**  Copyright (c) 2002-05 Rhett M. Hollander <rhett@alasir.com>
**
**  All rights reserved.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include "defines.h"

extern UTL intwr(UTL blksize, UTL passnum);
extern UTL intrd(UTL blksize, UTL passnum);
extern UTL floatwr(UTL blksize, UTL passnum);
extern UTL floatrd(UTL blksize, UTL passnum);
extern UTL intcp(UTL maxblk, UTL passnum);
extern UTL intsc(UTL maxblk, UTL passnum);
extern UTL intad(UTL maxblk, UTL passnum);
extern UTL inttr(UTL maxblk, UTL passnum);
extern UTL floatcp(UTL maxblk, UTL passnum);
extern UTL floatsc(UTL maxblk, UTL passnum);
extern UTL floatad(UTL maxblk, UTL passnum);
extern UTL floattr(UTL maxblk, UTL passnum);

#if (I386_ASM) || (I386_ASM_MF) || (AMD64_ASM) || (AMD64_ASM_MF)

#if (I386_ASM) || (I386_ASM_MF)

#if (FreeBSD)
#include <fcntl.h>
#elif (NetBSD) || (OpenBSD)
#include <sys/types.h>
#include <machine/sysarch.h>
#elif (Linux)
#include <sys/io.h>
#endif

extern U32 cpuinfo(U32 mode);

#endif

extern U32 mmxwr(U32 blksize, U32 passnum);
extern U32 mmxrd(U32 blksize, U32 passnum);
extern U32 mmxcp(U32 maxblk, U32 passnum);
extern U32 mmxsc(U32 maxblk, U32 passnum);
extern U32 mmxad(U32 maxblk, U32 passnum);
extern U32 mmxtr(U32 maxblk, U32 passnum);
extern U32 ssewr(U32 blksize, U32 passnum);
extern U32 sserd(U32 blksize, U32 passnum);
extern U32 ssecp(U32 maxblk, U32 passnum);
extern U32 ssesc(U32 maxblk, U32 passnum);
extern U32 ssead(U32 maxblk, U32 passnum);
extern U32 ssetr(U32 maxblk, U32 passnum);

#endif

void help(void) {

#if (I386_ASM) || (I386_ASM_MF)
    S32 ret;
#endif

    printf("\n%s\n%s\n%s\n%s\n%s\n",
      "USAGE: ramspeed -b ID [-g size] [-m size] [-l runs]",
      "-b  runs a specified benchmark (by an ID number):",
      "    1 -- INTmark (writing)      4 -- FLOATmark (writing)",
      "    2 -- INTmark (reading)      5 -- FLOATmark (reading)",
      "    3 -- INTmem                 6 -- FLOATmem");
#if (I386_ASM) || (I386_ASM_MF) || (AMD64_ASM) || (AMD64_ASM_MF)
    printf("%s\n%s\n%s\n",
      "    7 -- MMXmark (writing)     10 -- SSEmark (writing)",
      "    8 -- MMXmark (reading)     11 -- SSEmark (reading)",
      "    9 -- MMXmem                12 -- SSEmem");
#endif
    printf("%s\n%s\n%s\n%s\n",
      "-g  specifies a # of Gbytes per pass (default is 4)",
      "-m  specifies a # of Mbytes per array (default is 16)",
      "-l  enables the BatchRun mode (for *mem benchmarks only),",
      "    and specifies a # of runs (suggested is 5)");

#if (I386_ASM) || (I386_ASM_MF)

    printf("\n");

/* the following OS-dependent calls will fail without super-user's privileges */

#if (FreeBSD)    
    ret = open("/dev/io", O_RDWR);
    if(ret > 0) {
    	/* in verbose privileged mode */
    	cpuinfo(3);
	close(ret);
    } else {
	/* in verbose regular mode */
    	cpuinfo(1);
    }
#elif (NetBSD) || (OpenBSD)
    ret = i386_iopl(3);
    if(ret == 0) cpuinfo(3);
    else cpuinfo(1);
#elif (Linux)
    ret = iopl(3);
    if (ret == 0) cpuinfo(3);
    else cpuinfo(1);
#else
    /* ret is here just to shut up the compiler */
    ret = cpuinfo(1);
#endif

#endif

}

int main(int argc, char *argv[]) {
    UTL maxblk  = (MEMSIZE << 10);
    UTL passize = MEMPASS;
    U32 bid = 0, bnum = 1;
    F64 cplr = 0, adlr = 0, sclr = 0, trlr = 0;
    UTL blksize, passnum, time;
    STL opt, temp;
    U32 bmark, cnt;
    S32 ch;
    F64 cp, ad, sc, tr;
#if (I386_ASM) || (I386_ASM_MF)
    U32 cpucaps;
#endif

    printf("%s%s%s%s%s%s\n",
      "RAMspeed (", POINT_OS, ") v", RVERSION,
      " by Rhett M. Hollander and others, 2002-", RYEAR);

    while((ch = getopt(argc, argv, "b:g:m:l:")) != -1)
      switch(ch) {

        case 'b':
	    opt = atol(optarg);
	    if(opt < 0) opt = 0;
	    bid = (U32)opt;
	    break;

	case 'g':
	    opt = atol(optarg);
	    temp = 1;
	    passize = 0;
	    for(cnt = 0; cnt < (sizeof(STL) << 3); cnt++) {
		if(opt == temp) passize = opt;
		temp = (temp << 1);
	    }
	    if(!passize) {
		printf("\nERROR: invalid pass size\n");
		return(1);
	    }
	    break;

	case 'm':
	    opt = (atol(optarg) << 10);
	    temp = 1;
	    maxblk = 0;
	    for(cnt = 0; cnt < (sizeof(STL) << 3); cnt++) {
		if(opt == temp) maxblk = opt;
		temp = (temp << 1);
	    }
	    if(!maxblk) {
		printf("\nERROR: invalid array size\n");
		return(1);
	    }
	    break;

	case 'l':
	    opt = atol(optarg);
	    if(opt < 1) opt = 1;
	    bnum = (U32)opt;
	    break;

	default:
	    help();
	    return(1);

    }

    if(!bid) {
	help();
	return(1);
    }

#if (I386_ASM) || (I386_ASM_MF)
    /* be silent */
    cpucaps = cpuinfo(0);
#endif

    /* in fact, per pass values are only true for *mark benchmarks;
     * for *mem, cp & sc use 2 times, ad & tr -- 3 times more */
    printf("\n%luGb per pass mode\n\n", passize);

    passnum = (passize << 20)/maxblk;

    switch(bid) {

	case(1):
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = intwr(blksize, passnum);
		printf("INTEGER & WRITING  %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(2):
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = intrd(blksize, passnum);
		printf("INTEGER & READING  %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(3):
	    if(bnum > 1) printf("%u-benchmark INTmem BatchRun mode\n\n", bnum);
	    for(bmark = 0; bmark < bnum; bmark++) {
		if(bnum > 1) printf("Benchmark #%u:\n", bmark+1);
		time = intcp(maxblk, passnum);
		cp = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("INTEGER   Copy:      %.2f Mb/s\n", cp);
		time = intsc(maxblk, passnum);
		sc = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("INTEGER   Scale:     %.2f Mb/s\n", sc);
		time = intad(maxblk, passnum);
		ad = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("INTEGER   Add:       %.2f Mb/s\n", ad);
		time = inttr(maxblk, passnum);
		tr = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("INTEGER   Triad:     %.2f Mb/s\n", tr);
		printf("---\nINTEGER   AVERAGE:   %.2f Mb/s\n\n",
		  (cp + sc + ad + tr)/4.0);
		if(bnum > 1) {
		    cplr += cp; sclr += sc; adlr += ad; trlr += tr;
		}
	    }
	    if(bnum > 1) {
		printf("INTEGER BatchRun   Copy:      %.2f Mb/s\n", cplr/(F64)bnum);
		printf("INTEGER BatchRun   Scale:     %.2f Mb/s\n", sclr/(F64)bnum);
		printf("INTEGER BatchRun   Add:       %.2f Mb/s\n", adlr/(F64)bnum);
		printf("INTEGER BatchRun   Triad:     %.2f Mb/s\n", trlr/(F64)bnum);
		printf("---\nINTEGER BatchRun   AVERAGE:   %.2f Mb/s\n\n",
		  (cplr + sclr + adlr + trlr)/(F64)(bnum << 2));
	    }
	    break;

	case(4):
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = floatwr(blksize, passnum);
		printf("FL-POINT & WRITING %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(5):   
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = floatrd(blksize, passnum);
		printf("FL-POINT & READING %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(6):
	    if(bnum > 1) printf("%u-benchmark FLOATmem BatchRun mode\n\n", bnum);
	    else bnum = 1;
	    for(bmark = 0; bmark < bnum; bmark++) {
		if(bnum > 1) printf("Benchmark #%u:\n", bmark+1);
		time = floatcp(maxblk, passnum);
		cp = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("FL-POINT  Copy:      %.2f Mb/s\n", cp);
		time = floatsc(maxblk, passnum);
		sc = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("FL-POINT  Scale:     %.2f Mb/s\n", sc);
		time = floatad(maxblk, passnum);
		ad = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("FL-POINT  Add:       %.2f Mb/s\n", ad);
		time = floattr(maxblk, passnum);
		tr = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("FL-POINT  Triad:     %.2f Mb/s\n", tr);
		printf("---\nFL-POINT  AVERAGE:   %.2f Mb/s\n\n",
		  (cp + sc + ad + tr)/4.0);
		if(bnum > 1) {
		    cplr += cp; sclr += sc; adlr += ad; trlr += tr;
		}
	    }
	    if(bnum > 1) {
		printf("FL-POINT BatchRun  Copy:      %.2f Mb/s\n", cplr/(F64)bnum);
		printf("FL-POINT BatchRun  Scale:     %.2f Mb/s\n", sclr/(F64)bnum);
		printf("FL-POINT BatchRun  Add:       %.2f Mb/s\n", adlr/(F64)bnum);
		printf("FL-POINT BatchRun  Triad:     %.2f Mb/s\n", trlr/(F64)bnum);
		printf("---\nFL-POINT BatchRun  AVERAGE:   %.2f Mb/s\n\n",
		  (cplr + sclr + adlr + trlr)/(F64)(bnum << 2));
	    }
	    break;

#if (I386_ASM) || (I386_ASM_MF) || (AMD64_ASM) || (AMD64_ASM_MF)
	case(7):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x4)) {
		printf("ERROR: no MMX support available\n");
		return(1);
	    }
#endif
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = mmxwr(blksize, passnum);
		printf("MMX & WRITING      %8lu Kb block: %.2f Mb/s\n", blksize,
          	  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(8):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x4)) {
		printf("ERROR: no MMX support available\n");
		return(1);
	    }
#endif
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = mmxrd(blksize, passnum);
		printf("MMX & READING      %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(9):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x4)) {
		printf("ERROR: no MMX support available\n");
		return(1);
	    }
#endif
	    if(bnum > 1) printf("%u-benchmark MMXmem BatchRun mode\n\n", bnum);
	    else bnum = 1;
	    for(bmark = 0; bmark < bnum; bmark++) {
		if(bnum > 1) printf("Benchmark #%u:\n", bmark+1);
		time = mmxcp(maxblk, passnum);
		cp = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("MMX       Copy:      %.2f Mb/s\n", cp);
		time = mmxsc(maxblk, passnum);
		sc = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("MMX       Scale:     %.2f Mb/s\n", sc);
		time = mmxad(maxblk, passnum);
		ad = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("MMX       Add:       %.2f Mb/s\n", ad);
		time = mmxtr(maxblk, passnum);
		tr = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("MMX       Triad:     %.2f Mb/s\n", tr);
		printf("---\nMMX       AVERAGE:   %.2f Mb/s\n\n",
		  (cp + sc + ad + tr)/4.0);
		if(bnum > 1) {
		    cplr += cp; sclr += sc; adlr += ad; trlr += tr;
		}
	    }
	    if(bnum > 1) {
		printf("MMX BatchRun       Copy:      %.2f Mb/s\n", cplr/(F64)bnum);
		printf("MMX BatchRun       Scale:     %.2f Mb/s\n", sclr/(F64)bnum);
		printf("MMX BatchRun       Add:       %.2f Mb/s\n", adlr/(F64)bnum);
		printf("MMX BatchRun       Triad:     %.2f Mb/s\n", trlr/(F64)bnum);
		printf("---\nMMX BatchRun       AVERAGE:   %.2f Mb/s\n\n",
		  (cplr + sclr + adlr + trlr)/(F64)(bnum << 2));
	    }
	    break;

	case(10):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x40)) {
		printf("ERROR: no SSE support available\n");
		return(1);
	    }
#endif
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = ssewr(blksize, passnum);
		printf("SSE & WRITING      %8lu Kb block: %.2f Mb/s\n", blksize,
          	  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(11):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x40)) {
		printf("ERROR: no SSE support available\n");
		return(1);
	    }
#endif
	    for(blksize = STARTBLK; blksize <= maxblk; blksize = (blksize << 1)) {
		passnum = (passize << 20)/blksize;
		time = sserd(blksize, passnum);
		printf("SSE & READING      %8lu Kb block: %.2f Mb/s\n", blksize,
		  (F64)(passize << 10)*1000000.0/(F64)time);
	    }
	    break;

	case(12):
#if (I386_ASM) || (I386_ASM_MF)
	    if(!(cpucaps & 0x40)) {
		printf("ERROR: no SSE support available\n");
		return(1);
	    }
#endif
	    if(bnum > 1) printf("%u-benchmark SSEmem BatchRun mode\n\n", bnum);
	    else bnum = 1;
	    for(bmark = 0; bmark < bnum; bmark++) {
		if(bnum > 1) printf("Benchmark #%u:\n", bmark+1);
		time = ssecp(maxblk, passnum);
		cp = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("SSE       Copy:      %.2f Mb/s\n", cp);
		time = ssesc(maxblk, passnum);
		sc = (F64)(passize << 11)*1000000.0/(F64)time;
		printf("SSE       Scale:     %.2f Mb/s\n", sc);
		time = ssead(maxblk, passnum);
		ad = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("SSE       Add:       %.2f Mb/s\n", ad);
		time = ssetr(maxblk, passnum);
		tr = (F64)(passize*3072)*1000000.0/(F64)time;
		printf("SSE       Triad:     %.2f Mb/s\n", tr);
		printf("---\nSSE       AVERAGE:   %.2f Mb/s\n\n",
		  (cp + sc + ad + tr)/4.0);
		if(bnum > 1) {
		    cplr += cp; sclr += sc; adlr += ad; trlr += tr;
		}
	    }
	    if(bnum > 1) {
		printf("SSE BatchRun       Copy:      %.2f Mb/s\n", cplr/(F64)bnum);
		printf("SSE BatchRun       Scale:     %.2f Mb/s\n", sclr/(F64)bnum);
		printf("SSE BatchRun       Add:       %.2f Mb/s\n", adlr/(F64)bnum);
		printf("SSE BatchRun       Triad:     %.2f Mb/s\n", trlr/(F64)bnum);
		printf("---\nSSE BatchRun       AVERAGE:   %.2f Mb/s\n\n",
		  (cplr + sclr + adlr + trlr)/(F64)(bnum << 2));
	    }
	    break;
#endif

	default:
	    help();
	    printf("\nERROR: unknown benchmark ID\n");
	    return(1);
    }

    return(0);
}
